#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import os

from muninn.schema import *
from muninn.util import split_path
from muninn.remote import *


class ArchivePath(Text):
    _alias = "archive_path"

    @classmethod
    def validate(cls, value):
        super(ArchivePath, cls).validate(value)
        if os.path.isabs(value) or os.path.pardir in split_path(value):
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class Basename(Text):
    @classmethod
    def validate(cls, value):
        super(Basename, cls).validate(value)
        if value != os.path.basename(value):
            raise ValueError("invalid value %r for type %r" % (value, cls.name()))


class Remote(Text):
    @classmethod
    def validate(cls, value):
        super(Remote, cls).validate(value)
        # NOTE: We used to verify the value started with one of the supported prefixes
        # But now the list can be extended through a remote backend plugin, which is
        # archive dependent. The validation is now done at pull-time.


class Core(Mapping):
    uuid = UUID()
    active = Boolean(index=True)
    hash = Text(optional=True, index=True)
    size = Long(optional=True, index=True)
    metadata_date = Timestamp(index=True)
    archive_date = Timestamp(optional=True, index=True)
    archive_path = ArchivePath(optional=True)
    product_type = Text(index=True)
    product_name = Text(index=True)
    physical_name = Basename(index=True)
    validity_start = Timestamp(optional=True, index=True)
    validity_stop = Timestamp(optional=True, index=True)
    creation_date = Timestamp(optional=True, index=True)
    footprint = Geometry(optional=True, index=True)
    remote_url = Remote(optional=True)



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

from muninn.exceptions import Error


class Struct(object):
    def __init__(self, data=None, _depth=0):
        super(Struct, self).__init__()
        if data is not None:
            for key in data:
                if _depth == 0 and isinstance(data[key], dict):
                    self[key] = Struct(data[key], _depth=_depth+1)
                else:
                    self[key] = data[key]

    def __getitem__(self, key):
        try:
            return getattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __delitem__(self, key):
        try:
            return delattr(self, key)
        except AttributeError:
            raise KeyError(key)

    def __contains__(self, key):
        return hasattr(self, key)

    def __len__(self):
        return len(vars(self))

    def __iter__(self):
        return iter(vars(self))

    def __repr__(self):
        return "Struct(%r)" % vars(self)

    def update(self, other):
        '''update a struct, using the same semantics as dict.update()'''
        for key in other:
            other_item = other[key]
            if isinstance(other_item, Struct):
                if key not in self:
                    self[key] = Struct()
                else:
                    if not isinstance(self[key], Struct):
                        raise Error('Incompatible structs: %s vs %s' % (self, other))
                self[key].update(other_item)
            else:
                self[key] = other_item



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import muninn

from muninn.tools.utils import Processor, create_parser, parse_args_and_run


class StripProcessor(Processor):
    def __init__(self, args):
        super(StripProcessor, self).__init__(args)
        self.args = args

    def perform_operation(self, archive, product):
        archive.strip(product.core.uuid, force=self.args.force, cascade=False)
        return 1


def strip(args):
    processor = StripProcessor(args)
    with muninn.open(args.archive) as archive:
        products = archive.search(where=args.expression, property_names=['uuid'])
        returncode = processor.process(archive, args, products)
        archive.cleanup_derived_products()
        return returncode


def main():
    parser = create_parser(description="Strip products contained in a muninn archive (i.e. remove "
                           "products from disk, but don't remove the corresponding entries from the product "
                           "catalogue)", parallel=True)
    parser.add_argument("-f", "--force", action="store_true", help="also strip partially ingested products; note"
                        " that this can cause product files to be removed while in the process of being ingested")
    parser.add_argument("archive", metavar="ARCHIVE", help="identifier of the archive to use")
    parser.add_argument("expression", metavar="EXPRESSION", help="expression used to search for products to remove")
    return parse_args_and_run(parser, strip)


if __name__ == '__main__':
    main()



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

from muninn._compat import input
import muninn

from muninn.tools.utils import create_parser, parse_args_and_run


def ask_yes_no(question, default=True):
    prompt = "[y/n]" if default is None else "[Y/n]" if default else "[y/N]"
    while True:
        print(question + " " + prompt, end="")
        answer = input().lower()

        if default is not None and not answer:
            return default

        if answer in ("y", "yes", "n", "no"):
            return answer.startswith("y")

        print("")
        print("Please respond with \"yes\" or \"no\" (or \"y\" or \"n\").")


def destroy(args):
    with muninn.open(args.archive) as archive:
        if not args.yes:
            if args.catalogue_only:
                print(("You are about to remove the catalogue database for the archive \"%s\". "
                       "This operation cannot be undone!") % args.archive)
            else:
                print(("You are about to completely remove the archive \"%s\". "
                       "This operation cannot be undone!") % args.archive)
            if not ask_yes_no("Do you want to continue?", False):
                return 1

        if args.catalogue_only:
            archive.destroy_catalogue()
        else:
            archive.destroy()
    return 0


def main():
    parser = create_parser(description="Remove a muninn archive and its contents")
    parser.add_argument("-c", "--catalogue-only", action="store_true", help="only remove the catalogue database, "
                        "without touching (or removing anything from) the archive root path on disk")
    parser.add_argument("-y", "--yes", action="store_true", help="assume yes and do not prompt for confirmation")
    parser.add_argument("archive", metavar="ARCHIVE", help="identifier of the archive to use")
    return parse_args_and_run(parser, destroy)


if __name__ == '__main__':
    main()



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import argparse
import logging
import os
import re

import muninn

from muninn.tools.utils import Processor, create_parser, parse_args_and_run


class ExportProcessor(Processor):
    def __init__(self, args, target_path):
        super(ExportProcessor, self).__init__(args)
        self.args = args
        self.target_path = target_path

    def perform_operation(self, archive, product):
        archive.export(where=product.core.uuid, target_path=self.target_path, format=self.args.format)
        return 1


def export(args):
    target_path = os.getcwd() if args.directory is None else args.directory
    processor = ExportProcessor(args, target_path)

    with muninn.open(args.archive) as archive:
        if args.list_formats:
            if not archive.export_formats():
                print("no alternative export formats available")
            else:
                print("alternative export formats: " + " ".join(archive.export_formats()))
            print("")
            return 0

        if args.expression is None:
            logging.error("no search expression specified")
            return 1

        products = archive.search(where=args.expression, property_names=['uuid'])
        return processor.process(archive, args, products)
    return 0


def export_format(text):
    if re.match("[a-zA-Z]\\w*$", text) is None:
        raise argparse.ArgumentTypeError("invalid export format: %r" % text)

    return text


def directory(text):
    if not os.path.isdir(text):
        raise argparse.ArgumentTypeError("no such directory: %r" % text)

    return text


def main():
    parser = create_parser(description="Export products from a muninn archive.", parallel=True)
    parser.add_argument("-d", "--directory", type=directory, help="directory in which retrieved products will be"
                        " stored; by default, retrieved products will be stored in the current working directory")
    parser.add_argument("-f", "--format", type=export_format, help="format in which to export the products; if left"
                        " unspecified, the default export format for the product type will be used")
    parser.add_argument("-l", "--list-formats", action="store_true", help="list alternative (non-default) export"
                        " formats supported by the archive and exit")
    parser.add_argument("archive", metavar="ARCHIVE", help="identifier of the archive to use")
    parser.add_argument("expression", metavar="EXPRESSION", help="expression used to search for products to export")
    return parse_args_and_run(parser, export)


if __name__ == '__main__':
    main()



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import math
try:
    from collections.abc import MutableSequence
except ImportError:
    from collections import MutableSequence

from muninn.exceptions import Error


def polygon_rotation(pts):
    # return wether polygon is:
    #  1: anti-clockwise rotation (right-hand-rule) -> use inner area
    #  0: no rotation -> polygon is empty or invalid
    # -1: clockwise rotation (left-hand-rule) -> use outer area
    # this can be calculated by summing the outer products of consecutive pts (taking vectors from (0,0) to the pt)
    prev_pt = pts[0]
    sum = 0
    for pt in pts[1:]:
        sum += pt[1] * prev_pt[0] - pt[0] * prev_pt[1]
        prev_pt = pt
    if sum == 0:
        return 0
    return math.copysign(1, sum)


class Geometry(object):
    @property
    def min_x(self):
        return 0

    @property
    def max_x(self):
        return 0

    @property
    def min_y(self):
        return 0

    @property
    def max_y(self):
        return 0

    @staticmethod
    def from_geojson(geojson):
        type_ = geojson['type']
        coordinates = geojson['coordinates']

        if type_ == 'Point':
            return as_point(coordinates)

        elif type_ == 'LineString':
            return as_line_string(coordinates)

        elif type_ == 'Polygon':
            return as_polygon(coordinates)

        elif type_ == 'MultiPoint':
            return as_multi_point(coordinates)

        elif type_ == 'MultiLineString':
            return as_multi_line_string(coordinates)

        elif type_ == 'MultiPolygon':
            return as_multi_polygon(coordinates)

        else:
            raise Error('cannot convert geojson type: %s' % type_)

    def wrap(self):
        """
        Convert the geometry from one on a sphere to one that fits on a 2D lat/lon canvas with
        -90 <= latitude <= 90 and -180 <= longitude <= 180.
        This involves splitting geometries around the dateline and dealing with geometries that cover the poles
        """
        return self


class Point(Geometry):
    def __init__(self, x, y):
        self._coordinates = [x, y]

    @property
    def x(self):
        return self._coordinates[0]

    @x.setter
    def x(self, value):
        self._coordinates[0] = value

    @property
    def y(self):
        return self._coordinates[1]

    @y.setter
    def y(self, value):
        self._coordinates[1] = value

    @property
    def min_x(self):
        return self.x

    @property
    def max_x(self):
        return self.x

    @property
    def min_y(self):
        return self.y

    @property
    def max_y(self):
        return self.y

    @property
    def longitude(self):
        return self._coordinates[0]

    @longitude.setter
    def longitude(self, value):
        self._coordinates[0] = value

    @property
    def latitude(self):
        return self._coordinates[1]

    @latitude.setter
    def latitude(self, value):
        self._coordinates[1] = value

    def as_geojson(self):
        return {
            'type': 'Point',
            'coordinates': [self.x, self.y],
        }

    def as_wkt(self, tagged=True):
        wkt = "(%f %f)" % (self.x, self.y)
        return "POINT " + wkt if tagged else wkt

    def wrap(self):
        """
        This function assumes that the longitude is already in the range [-360,360].
        """
        # map lon to [-180, 180]
        lon = self.longitude
        return Point(lon + 360 if lon < -180 else (lon - 360 if lon > 180 else lon), self.latitude)

    def __getitem__(self, index):
        return self._coordinates[index]

    def __setitem__(self, index, value):
        self._coordinates[index] = value

    def __len__(self):
        return 2

    def __eq__(self, other):
        return self._coordinates == other._coordinates

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return "Point(x=%f, y=%f)" % (self.x, self.y)

    def __str__(self):
        return self.as_wkt()


class GeometrySequence(Geometry, MutableSequence):
    def __init__(self, geometries=[]):
        self._geometries = list(geometries)

    @property
    def min_x(self):
        return min([g.min_x for g in self._geometries])

    @property
    def max_x(self):
        return max([g.max_x for g in self._geometries])

    @property
    def min_y(self):
        return min([g.min_y for g in self._geometries])

    @property
    def max_y(self):
        return max([g.max_y for g in self._geometries])

    def insert(self, index, geometry):
        self._geometries.insert(index, geometry)

    def wrap(self):
        return self.__class__([geometry.wrap() for geometry in self._geometries])

    def __getitem__(self, index):
        return self._geometries[index]

    def __setitem__(self, index, point):
        self._geometries[index] = point

    def __delitem__(self, index):
        del self._geometries[index]

    def __len__(self):
        return len(self._geometries)

    def __eq__(self, other):
        return self._geometries == other._geometries

    def __ne__(self, other):
        return not self == other

    def __repr__(self):
        return "GeometrySequence(geometries=%r)" % self._geometries


class LineString(GeometrySequence):
    def __init__(self, geometries=[]):
        super(LineString, self).__init__(geometries)

        if len(self) == 1:
            raise Error("line string should be empty or should contain >= 2 points")

    def point(self, index):
        return self[index]

    def is_closed(self):
        return not self or self[0] == self[-1]

    def as_geojson(self):
        return {
            'type': 'LineString',
            'coordinates': [[point.x, point.y] for point in self],
        }

    def as_wkt(self, tagged=True):
        wkt = "(" + ", ".join(["%f %f" % (point.x, point.y) for point in self]) + ")" if self else "EMPTY"
        return "LINESTRING " + wkt if tagged else wkt

    def wrap(self):
        """
        Lines will be divided into sub-lines if they cross the dateline
        """
        lon, lat = self[0].wrap()
        prev_lon, prev_lat = lon, lat
        pts = [Point(lon, lat)]
        pts_set = [pts]
        for point in self[1:]:
            lon, lat = point.wrap()
            # rel_lon = lon mapped to [prev_lon - 180, prev_lon + 180]
            rel_lon = lon + 360 if lon < prev_lon - 180 else (lon - 360 if lon > prev_lon + 180 else lon)
            if rel_lon < -180:
                # crossing the dateline meridian -> split line
                mid_lat = lat + ((-180 - rel_lon) / (prev_lon - rel_lon)) * (prev_lat - lat)
                pts.append(Point(-180, mid_lat))
                pts = [Point(180, mid_lat)]
                pts_set.append(pts)
            elif rel_lon > 180:
                # crossing the dateline meridian -> split line
                mid_lat = prev_lat + ((180 - prev_lon) / (rel_lon - prev_lon)) * (lat - prev_lat)
                pts.append(Point(180, mid_lat))
                pts = [Point(-180, mid_lat)]
                pts_set.append(pts)
            prev_lon, prev_lat = lon, lat
            pts.append(Point(lon, lat))
        if len(pts_set) > 1:
            return MultiLineString([LineString(pts) for pts in pts_set])
        else:
            return LineString(pts)

    def __repr__(self):
        return "LineString(points=%r)" % self._geometries

    def __str__(self):
        return self.as_wkt()


class LinearRing(GeometrySequence):
    def __init__(self, geometries=[]):
        super(LinearRing, self).__init__(geometries)

        if self and len(self) < 3:
            raise ValueError("linear ring should be empty or should contain >= 3 points")

    def point(self, index):
        return self[index]

    def as_wkt(self, tagged=True):
        if not self:
            wkt = "EMPTY"
        else:
            wkt = "(%s, %f %f)" % (", ".join(["%f %f" % (point.x, point.y) for point in self]), self.point(0).x,
                                   self.point(0).y)
        return "LINESTRING " + wkt if tagged else wkt

    def __repr__(self):
        return "LinearRing(points=%r)" % self._geometries

    def __str__(self):
        return self.as_wkt()


class Polygon(GeometrySequence):
    def ring(self, index):
        return self[index]

    def exterior_ring(self):
        return self[0]

    def interior_ring(self, index):
        return self[index + 1]

    def as_wkt(self, tagged=True):
        wkt = "(" + ", ".join([geometry.as_wkt(False) for geometry in self]) + ")" if self else "EMPTY"
        return "POLYGON " + wkt if tagged else wkt

    def as_geojson(self):
        return {
            'type': 'Polygon',
            'coordinates': [[[point.x, point.y] for point in ring] for ring in self],
        }

    def wrap(self):
        """
        Polygons need to be split at the dateline. This also requires using a special 'unfolding' to make a polygon
        that covers the North and/or South pole to still cover the whole polar region on a flat 2D area.

        Any exclusion regions will be remove in the conversion.

        The special situation where a polygon covers both poles _and_ runs along the dateline will result in a single
        polygon with a wrong rotation. This type of polygon is turned into a geometry with a hole (i.e. outer polygon
        is the full earth bounding box, and the original polygon becomes the exclusion area).
        Input polygons should be properly oriented using the right-hand rule (= anti clockwise) or they may otherwise
        be turned into exclusions by this algorithm.
        """
        # We only wrap the outer ring
        ring = self.exterior_ring()
        lon, lat = ring[0].wrap()
        # current_area = {-1: lon < -180, 0: -180 <= lon <= 180, 1: lon >= 180}
        current_area = 0
        prev_lon, prev_lat = lon, lat
        pts = [Point(lon, lat)]
        pts_set = [pts]
        crossing_lat = []
        for point in ring[1:]:
            lon, lat = point.wrap()
            # rel_lon = lon mapped to [prev_lon - 180, prev_lon + 180]
            rel_lon = lon + 360 if lon < prev_lon - 180 else (lon - 360 if lon > prev_lon + 180 else lon)
            if rel_lon < -180:
                if current_area == -1:
                    # unsupported polygon
                    return self
                # crossing the dateline meridian -> split polygon
                mid_lat = lat + ((-180 - rel_lon) / (prev_lon - rel_lon)) * (prev_lat - lat)
                crossing_lat.append(mid_lat)
                pts.append(Point(-180, mid_lat))
                pts = [Point(180, mid_lat)]
                pts_set.append(pts)
                current_area -= 1
            elif rel_lon > 180:
                if current_area == 1:
                    # unsupported polygon
                    return self
                # crossing the dateline meridian -> split polygon
                mid_lat = prev_lat + ((180 - prev_lon) / (rel_lon - prev_lon)) * (lat - prev_lat)
                crossing_lat.append(mid_lat)
                pts.append(Point(180, mid_lat))
                pts = [Point(-180, mid_lat)]
                pts_set.append(pts)
                current_area += 1
            prev_lon, prev_lat = lon, lat
            pts.append(Point(lon, lat))
        if len(pts_set) == 1:
            assert len(crossing_lat) == 0
            if polygon_rotation(pts) < 0:
                world = LinearRing([Point(-180, -90), Point(180, -90), Point(180, 90), Point(-180, 90),
                                    Point(-180, -90)])
                return Polygon([world, LinearRing(pts)])
            else:
                return Polygon([LinearRing(pts)])
        # prepend final pts to first ring
        if pts[-1] == pts_set[0][0]:
            del pts[-1]
        pts.extend(pts_set[0])
        pts_set[0] = pts
        del pts_set[-1]
        # check if we need to connect via the north pole
        if len(crossing_lat) > 0:
            max_lat = max(crossing_lat)
            max_index = crossing_lat.index(max_lat)
            next_index = max_index + 1 if max_index < len(crossing_lat) - 1 else 0
            if pts_set[max_index][-1][0] > pts_set[next_index][0][0]:
                # connect pts via the north pole
                pts_set[max_index].append(Point(180, 90))
                pts_set[max_index].append(Point(-180, 90))
                if max_index != next_index:
                    pts_set[max_index].extend(pts_set[next_index])
                    pts_set[next_index] = pts_set[max_index]
                    del pts_set[max_index]
                    del crossing_lat[max_index]
        # check if we need to connect via the south pole
        if len(crossing_lat) > 0:
            min_lat = min(crossing_lat)
            min_index = crossing_lat.index(min_lat)
            next_index = min_index + 1 if min_index < len(crossing_lat) - 1 else 0
            if pts_set[min_index][-1][0] < pts_set[next_index][0][0]:
                # connect pts via the south pole
                pts_set[min_index].append(Point(-180, -90))
                pts_set[min_index].append(Point(180, -90))
                if min_index != next_index:
                    pts_set[min_index].extend(pts_set[next_index])
                    pts_set[next_index] = pts_set[min_index]
                    del pts_set[min_index]
                    del crossing_lat[min_index]
        for pts in pts_set:
            # close ring
            pts.append(pts[0])
        if len(pts_set) == 1:
            return Polygon([LinearRing(pts)])
        else:
            return MultiPolygon([Polygon([LinearRing(pts)]) for pts in pts_set])

    def __repr__(self):
        return "Polygon(rings=%r)" % self._geometries

    def __str__(self):
        return self.as_wkt()


class MultiPoint(GeometrySequence):
    def point(self, index):
        return self[index]

    def as_geojson(self):
        return {
            'type': 'MultiPoint',
            'coordinates': [[point.x, point.y] for point in self],
        }

    def as_wkt(self, tagged=True):
        wkt = "(" + ", ".join([geometry.as_wkt(False) for geometry in self]) + ")" if self else "EMPTY"
        return "MULTIPOINT " + wkt if tagged else wkt

    def __repr__(self):
        return "MultiPoint(points=%r)" % self._geometries

    def __str__(self):
        return self.as_wkt()


class MultiLineString(GeometrySequence):
    def line_string(self, index):
        return self[index]

    def as_geojson(self):
        return {
            'type': 'MultiLineString',
            'coordinates': [[[point.x, point.y] for point in linestring] for linestring in self],
        }

    def as_wkt(self, tagged=True):
        wkt = "(" + ", ".join([geometry.as_wkt(False) for geometry in self]) + ")" if self else "EMPTY"
        return "MULTILINESTRING " + wkt if tagged else wkt

    def wrap(self):
        lines = []
        for line in self._geometries:
            geometry = line.wrap()
            if type(geometry) is MultiLineString:
                lines.extend(geometry)
            else:
                lines.append(geometry)
        return MultiLineString(lines)

    def __repr__(self):
        return "MultiLineString(line_strings=%r)" % self._geometries

    def __str__(self):
        return self.as_wkt()


class MultiPolygon(GeometrySequence):
    def polygon(self, index):
        return self[index]

    def as_wkt(self, tagged=True):
        wkt = "(" + ", ".join([geometry.as_wkt(False) for geometry in self]) + ")" if self else "EMPTY"
        return "MULTIPOLYGON " + wkt if tagged else wkt

    def as_geojson(self):
        return {
            'type': 'MultiPolygon',
            'coordinates': [[[[point.x, point.y] for point in ring] for ring in polygon] for polygon in self],
        }

    def wrap(self):
        polys = []
        for poly in self._geometries:
            geometry = poly.wrap()
            if type(geometry) is MultiPolygon:
                polys.extend(geometry)
            else:
                polys.append(geometry)
        return MultiPolygon(polys)

    def __repr__(self):
        return "MultiPolygon(polygons=%r)" % self._geometries

    def __str__(self):
        return self.as_wkt()


def as_point(iterable):
    return Point(*iterable)


def as_line_string(iterable):
    return LineString(map(as_point, iterable))


def as_linear_ring(iterable):
    return LinearRing(map(as_point, iterable))


def as_polygon(iterable):
    return Polygon(map(as_linear_ring, iterable))


def as_multi_point(iterable):
    return MultiPoint(map(as_point, iterable))


def as_multi_line_string(iterable):
    return MultiLineString(map(as_line_string, iterable))


def as_multi_polygon(iterable):
    return MultiPolygon(map(as_polygon, iterable))



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import os
import argparse

import muninn

from muninn.tools.utils import Processor, create_parser, parse_args_and_run


def directory(text):
    if not os.path.isdir(text):
        raise argparse.ArgumentTypeError("no such directory: %r" % text)

    return text


class RetrieveProcessor(Processor):
    def __init__(self, args, target_path):
        super(RetrieveProcessor, self).__init__(args)
        self.args = args
        self.target_path = target_path

    def perform_operation(self, archive, product):
        archive.retrieve(product.core.uuid, target_path=self.target_path, use_symlinks=self.args.link,
                         verify_hash=self.args.verify_hash)
        return 1


def retrieve(args):
    target_path = os.getcwd() if args.directory is None else args.directory
    processor = RetrieveProcessor(args, target_path)

    with muninn.open(args.archive) as archive:
        products = archive.search(where=args.expression, property_names=['uuid'])
        return processor.process(archive, args, products)


def main():
    parser = create_parser(description="Retrieve products from a muninn archive.", parallel=True)
    parser.add_argument("-d", "--directory", type=directory, help="directory in which retrieved products will be"
                        " stored; by default, retrieved products will be stored in the current working directory")
    parser.add_argument("-l", "--link", action="store_true", help="retrieve using symbolic links instead of copy")
    parser.add_argument("--verify-hash", action="store_true",
                        help="verify the hash of the product after it has been retrieved")
    parser.add_argument("archive", metavar="ARCHIVE", help="identifier of the archive to use")
    parser.add_argument("expression", metavar="EXPRESSION", help="expression used to search for products to retrieve")
    return parse_args_and_run(parser, retrieve)


if __name__ == '__main__':
    main()



import os.path

import muninn.util as util


class StorageBackend(object):
    def __init__(self, tempdir=None):
        self.supports_symlinks = False
        self.global_prefix = ''

        if tempdir is not None:
            tmp_root = os.path.realpath(tempdir)
            util.make_path(tmp_root)
            self._tmp_root = tmp_root
        else:
            self._tmp_root = None

    def get_tmp_root(self, product):
        if self._tmp_root is not None:
            util.make_path(self._tmp_root)
            return self._tmp_root

    def run_for_product(self, product, fn, use_enclosing_directory):
        tmp_root = self.get_tmp_root(product)
        with util.TemporaryDirectory(dir=tmp_root, prefix=".run_for_product-",
                                     suffix="-%s" % product.core.uuid.hex) as tmp_path:
            self.get(product, tmp_path, use_enclosing_directory)
            paths = [os.path.join(tmp_path, basename) for basename in os.listdir(tmp_path)]
            return fn(paths)

    def prepare(self):  # pragma: no cover
        # Prepare storage for use.
        raise NotImplementedError()

    def exists(self):  # pragma: no cover
        # Check that storage exists.
        raise NotImplementedError()

    def initialize(self, configuration):  # pragma: no cover
        # Initialize storage.
        raise NotImplementedError()

    def destroy(self):  # pragma: no cover
        # Destroy storage
        raise NotImplementedError()

    def put(self, paths, properties, use_enclosing_directory, use_symlinks=None,
            retrieve_files=None, run_for_product=None):  # pragma: no cover
        # Place product file(s) into storage
        raise NotImplementedError()

    def get(self, product, target_path, use_enclosing_directory, use_symlinks=None):  # pragma: no cover
        # Retrieve product file(s) from storage
        raise NotImplementedError()

    def size(self, product_path):  # pragma: no cover
        # Return product storage size
        raise NotImplementedError()

    def delete(self, product_path, properties):  # pragma: no cover
        # Delete product file(s) from storage
        raise NotImplementedError()

    def move(self, product, archive_path, paths=None):  # pragma: no cover
        # Move product
        raise NotImplementedError()

    def current_archive_path(self, paths, properties):  # pragma: no cover
        raise NotImplementedError()



import json
import logging
import os

from .base import StorageBackend

from muninn.schema import Mapping, Text
from muninn.exceptions import Error, StorageError
import muninn.util as util
import muninn.config as config

import swiftclient

logging.getLogger("swiftclient").setLevel(logging.CRITICAL)


class _SwiftConfig(Mapping):
    _alias = "swift"

    authurl = Text()
    container = Text(optional=True)
    user = Text(optional=True)
    key = Text(optional=True)


def create(configuration, tempdir, auth_file):
    options = config.parse(configuration.get("swift", {}), _SwiftConfig)

    # if access_key and secret_access_key missing, use auth_file
    if 'user' not in options and 'key' not in options and auth_file is not None:
        credentials = json.loads(open(auth_file).read())
        for key, value in credentials.items():
            if key == options['authurl'] and value.get('auth_type') == 'Swift':
                for option in ('user', 'key', 'container'):
                    if option in value and option not in options:
                        options[option] = value[option]
                break

    # check that mandatory options are configured
    for option in ('user', 'key', 'container'):
        if option not in options:
            raise Error("'%s' not configured" % option)

    _SwiftConfig.validate(options)
    return SwiftStorageBackend(tempdir=tempdir, **options)


class SwiftStorageBackend(StorageBackend):  # TODO '/' in keys to indicate directory, 'dir/' with contents?
    def __init__(self, container, user, key, authurl, tempdir=None):
        super(SwiftStorageBackend, self).__init__(tempdir)

        self.container = container
        self._root = container

        self._conn = swiftclient.Connection(
            user=user,
            key=key,
            authurl=authurl
        )

    def _object_keys(self, product_path):
        product_path = product_path.replace('\\', '/')
        sub_objects = self._conn.get_container(self.container, prefix=product_path)[1]
        return [sub_object['name'] for sub_object in sub_objects]

    def prepare(self):
        if not self.exists():
            self._conn.put_container(self.container)

    def exists(self):
        try:
            self._conn.get_container(self.container)
            return True
        except swiftclient.exceptions.ClientException as e:
            if e.http_status == 404:
                return False
            else:
                raise

    def destroy(self):  # TODO individually deleting objects
        if self.exists():
            for data in self._conn.get_container(self.container)[1]:
                self._conn.delete_object(self.container, data['name'])
            self._conn.delete_container(self.container)

    def current_archive_path(self, paths, properties):
        raise Error("Swift storage backend does not support ingesting already archived products")

    def put(self, paths, properties, use_enclosing_directory, use_symlinks=None,
            retrieve_files=None, run_for_product=None):

        if use_symlinks:
            raise Error("Swift storage backend does not support symlinks")

        anything_stored = False
        try:
            archive_path = properties.core.archive_path
            physical_name = properties.core.physical_name

            if not use_enclosing_directory and retrieve_files is None:
                assert len(paths) == 1 and os.path.basename(paths[0]) == physical_name

            tmp_root = self.get_tmp_root(properties)
            with util.TemporaryDirectory(dir=tmp_root, prefix=".put-",
                                         suffix="-%s" % properties.core.uuid.hex) as tmp_path:
                if retrieve_files:
                    paths = retrieve_files(tmp_path)

                # Upload file(s)
                for path in paths:
                    key = util.fwd_join(archive_path, physical_name)

                    # Add enclosing dir
                    if use_enclosing_directory:
                        key = util.fwd_join(key, os.path.basename(path))

                    if os.path.isdir(path):
                        self._conn.put_object(self.container, key+'/', contents=b'')
                        anything_stored = True

                        for root, subdirs, files in os.walk(path):
                            rel_root = os.path.relpath(root, path)

                            for subdir in subdirs:
                                dirkey = os.path.normpath(os.path.join(key, rel_root, subdir))+'/'
                                dirkey = dirkey.replace('\\', '/')
                                self._conn.put_object(self.container, dirkey, contents=b'')
                                anything_stored = True

                            for filename in files:
                                filekey = os.path.normpath(os.path.join(key, rel_root, filename))
                                filekey = filekey.replace('\\', '/')
                                filepath = util.fwd_join(root, filename)
                                with open(filepath, 'rb') as f:
                                    self._conn.put_object(self.container, filekey, contents=f.read())
                                    anything_stored = True
                    else:
                        with open(path, 'rb') as f:
                            self._conn.put_object(self.container, key, contents=f.read())
                            anything_stored = True

                if run_for_product is not None:
                    run_for_product(paths)

        except Exception as e:
            raise StorageError(e, anything_stored)

    def get(self, product, target_path, use_enclosing_directory, use_symlinks=None):
        paths = set()

        if use_symlinks:
            raise Error("Swift storage backend does not support symlinks")

        archive_path = product.core.archive_path
        product_path = util.fwd_join(archive_path, product.core.physical_name)

        keys = self._object_keys(product_path)
        if not keys:
            raise Error("no data for product '%s' (%s)" % (product.core.product_name, product.core.uuid))

        for key in keys:
            rel_path = os.path.relpath(key, archive_path).replace('\\', '/')
            if use_enclosing_directory:
                rel_path = '/'.join(rel_path.split('/')[1:])
            paths.add(os.path.normpath(os.path.join(target_path, rel_path.split('/')[0])))
            target = os.path.normpath(os.path.join(target_path, rel_path))
            if key.endswith('/'):
                util.make_path(target)
            else:
                dirname = os.path.dirname(target)
                if dirname != '':
                    util.make_path(dirname)
                binary = self._conn.get_object(self.container, key)[1]
                with open(target, 'wb') as f:
                    f.write(binary)

        return list(paths)

    def delete(self, product_path, properties):
        for key in self._object_keys(product_path):
            self._conn.delete_object(self.container, key)

    def size(self, product_path):
        total = 0
        for data in self._conn.get_container(self.container, prefix=product_path)[1]:
            total += data['bytes']
        return total

    def move(self, product, archive_path, paths=None):
        # Ignore if product already there
        if product.core.archive_path == archive_path:
            return paths

        product_path = util.fwd_join(product.core.archive_path, product.core.physical_name)
        new_product_path = util.fwd_join(archive_path, product.core.physical_name)

        keys = self._object_keys(product_path)
        if not keys:
            raise Error("no data for product '%s' (%s)" % (product.core.product_name, product.core.uuid))

        for key in keys:
            new_key = os.path.normpath(os.path.join(new_product_path, os.path.relpath(key, product_path)))
            new_key = new_key.replace('\\', '/')
            if key.endswith('/'):
                self._conn.put_object(self.container, new_key+'/', contents=b'')
            else:
                self._conn.copy_object(self.container, key, util.fwd_join(self.container, new_key))

            self._conn.delete_object(self.container, key)

        return paths



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import re
import functools
import inspect
import json

try:
    import psycopg2
    import psycopg2.extensions
    import psycopg2.extras
except ImportError:
    pass

try:
    import pg8000
    pg8000.paramstyle = 'pyformat'
except ImportError:
    pass

from .base import DatabaseBackend

import muninn.config as config
import muninn.database.sql as sql
import muninn.database.ewkb as ewkb
import muninn.geometry as geometry

from muninn.exceptions import *
from muninn.function import Prototype
from muninn.schema import *
from muninn.struct import Struct

from muninn._compat import dictkeys, dictvalues, is_python2_unicode

PG_UNIQUE_VIOLATION = '23505'


def _get_db_type_id(connection, typename):
    try:
        cursor = connection.cursor()
        try:
            cursor.execute("SELECT NULL::%s" % typename.lower())
            if not cursor.description:
                raise InternalError("unable to retrieve type object id of database type: \"%s\"" % typename.upper())
            type_id = cursor.description[0][1]
        finally:
            cursor.close()
    except Exception:
        connection.rollback()
        raise
    else:
        connection.commit()

    return type_id


def geometry_recv(data, offset, length):  # TODO binary send/recv needed for pg8000 <= 1.15
    return ewkb.decode_ewkb(data[offset:offset+length])


def geometry_recv_hex(data):
    return ewkb.decode_hexewkb(data)


def geometry_send(geometry):
    return ewkb.encode_ewkb(geometry)


def geometry_send_hex(geometry):
    return ewkb.encode_hexewkb(geometry)


def _connect_pg8000(connection_string):
    kwargs = dict(arg.split('=') for arg in connection_string.split(' '))
    if 'dbname' in kwargs:
        kwargs['database'] = kwargs.pop('dbname')
    _connection = pg8000.connect(**kwargs)

    geography_oid = _get_db_type_id(_connection, "geography")

    if hasattr(_connection, 'register_in_adapter'):
        _connection.register_in_adapter(geography_oid, geometry_recv_hex)
    else:
        _connection.pg_types[geography_oid] = (pg8000.core.FC_BINARY, geometry_recv)

    for type_ in (
        geometry.Point,
        geometry.Polygon,
        geometry.LineString,
        geometry.MultiPoint,
        geometry.MultiPolygon,
        geometry.MultiLineString,
    ):
        if hasattr(_connection, 'register_out_adapter'):
            try:
                getargspec = inspect.getfullargspec
            except AttributeError:
                getargspec = inspect.getargspec
            # pg8000 removed the oid argument in v1.22
            if len(getargspec(_connection.register_out_adapter).args) == 3:
                _connection.register_out_adapter(type_, geometry_send_hex)
            else:
                _connection.register_out_adapter(type_, geography_oid, geometry_send_hex)
        else:
            _connection.py_types[type_] = (geography_oid, pg8000.core.FC_BINARY, geometry_send)

    return _connection


def _adapt_geometry(geometry):
    """Return the hexadecimal extended well known binary format (hexewkb) representation of a Geometry instance."""
    return psycopg2.extensions.AsIs("'%s'" % ewkb.encode_hexewkb(geometry))


def _cast_geography(hexewkb, cursor):
    """Construct a Geometry instance from its hexadecimal extended well known binary format (hexewkb) representation."""
    if hexewkb is None:
        return hexewkb
    return ewkb.decode_hexewkb(hexewkb)


def _connect_psycopg2(connection_string):
    _connection = psycopg2.connect(connection_string)

    # Register adapter and cast for the UUID type.
    psycopg2.extras.register_uuid(conn_or_curs=_connection)

    # Register adapter for the Geometry type.
    psycopg2.extensions.register_adapter(geometry.Geometry, _adapt_geometry)

    # Register cast for the Geometry type.
    geography_oid = _get_db_type_id(_connection, "geography")
    geography_type = psycopg2.extensions.new_type((geography_oid,), "GEOGRAPHY", _cast_geography)
    psycopg2.extensions.register_type(geography_type, _connection)

    return _connection


class _PostgresqlConfig(Mapping):
    _alias = "postgresql"

    library = Text(optional=True)
    connection_string = Text()
    table_prefix = Text(optional=True)


def create(configuration):
    options = config.parse(configuration.get("postgresql", {}), _PostgresqlConfig)
    _PostgresqlConfig.validate(options)
    return PostgresqlBackend(**options)


class PostgresqlError(Error):
    def __init__(self, message=None):
        message = "postgresql backend error" + ("" if not message else ": " + message)
        super(PostgresqlError, self).__init__(message)


def translate_errors(func):
    """Decorator that translates db 2.0 api exceptions into muninn exceptions."""
    @functools.wraps(func)
    def translate_errors_(self, *args, **kwargs):
        try:
            return func(self, *args, **kwargs)
        except self._connection._backend.Error as _error:
            message = None

            # psycopg2
            if self._library == 'psycopg2':
                try:
                    message = _error.diag.message_primary
                    if message is not None:
                        try:
                            message_detail = _error.diag.message_detail
                            if message_detail:
                                message += " [" + message_detail + "]"
                        except AttributeError:
                            pass
                except AttributeError:
                    pass

            elif self._library == 'pg8000':
                try:
                    message = _error.args[0]['M']
                except (TypeError, IndexError, AttributeError, KeyError):
                    pass

            # fallback
            if message is None:
                message = ' '.join(str(_error).split())

            raise PostgresqlError(message)

    return translate_errors_


# TODO use sub-classes for psycopg2, pg8000

class PostgresqlConnection(object):
    """Wrapper for a database connection that defers (re)connection until an attempt is made to use the
    connection.

    Only non-nested transactions are supported, no auto-commit or nested transactions. A transaction can be started
    using the context manager interface.

    """
    def __init__(self, connection_string, library):
        self._connection_string = connection_string
        self._library = library
        self._connection = None
        self._in_transaction = False

        if library == 'psycopg2':
            try:
                self._backend = psycopg2
            except NameError:
                raise Error('could not import psycopg2')

        elif library == 'pg8000':
            try:
                self._backend = pg8000
            except NameError:
                raise Error('could not import pg8000')
        else:
            raise Error('no such library: %s' % library)

    def __enter__(self):
        # Begin a transaction. The transaction is not started immediately, but a state change is recorded such that
        # attempts to start nested transactions can be detected. Also, the connection with the database is
        # re-established if necessary.
        #
        if self._in_transaction:
            raise InternalError("nested transactions are not supported")

        # Reconnect if necessary.
        if self._connection is None:
            self._connect()

        # Change state to guard against nested transactions.
        self._in_transaction = True

        return self

    def __exit__(self, type, value, traceback):
        # End a transaction, either by committing the transaction, or by rolling back in case an exception occured.
        try:
            if type is None:
                self._connection.commit()
            else:
                self._connection.rollback()
        finally:
            self._in_transaction = False
            self.close()

    def _connect(self):
        # Re-establish the connection to the database.
        if self._library == 'psycopg2':
            self._connection = _connect_psycopg2(self._connection_string)
        else:
            self._connection = _connect_pg8000(self._connection_string)

    def _disconnect(self):
        self._connection.close()
        self._connection = None

    def close(self):
        if self._in_transaction:
            raise InternalError("unable to close the connection with the database while a transaction is in progress")

        # Close the connection.
        if self._connection is not None:
            self._disconnect()

    def cursor(self):
        if not self._in_transaction:
            raise InternalError("creating a cursor requires an active transaction")

        return self._connection.cursor()

    @property
    def encoding(self):
        return self._connection.encoding


class PostgresqlBackend(DatabaseBackend):
    def __init__(self, connection_string="", table_prefix="", library="psycopg2"):
        self._connection = PostgresqlConnection(connection_string, library)
        self._library = library

        if table_prefix and not re.match(r"[a-z][_a-z]*(\.[a-z][_a-z]*)*", table_prefix):
            raise ValueError("invalid table_prefix %s" % table_prefix)
        self._table_prefix = table_prefix

        self._core_table_name = self._table_name("core")
        self._link_table_name = self._table_name("link")
        self._tag_table_name = self._table_name("tag")

        self._namespace_schemas = {}
        self._sql_builder = sql.SQLBuilder({}, sql.TypeMap(), {}, self._table_name, self._placeholder,
                                           self._placeholder, self._rewriter_property)

    def _create_tables_sql(self):
        result = []
        # Create the table for the core namespace.
        result.append(self._sql_builder.build_create_table_query("core"))
        result.append("ALTER TABLE %s ADD PRIMARY KEY (uuid);" % self._core_table_name)
        result.append("ALTER TABLE %s ADD CONSTRAINT %s_archive_path_uniq UNIQUE (archive_path, physical_name);" %
                      (self._core_table_name, self._core_table_name))
        result.append("ALTER TABLE %s ADD CONSTRAINT %s_product_name_uniq UNIQUE (product_type, product_name);" %
                      (self._core_table_name, self._core_table_name))

        schema = self._namespace_schema("core")
        for name in schema:
            if schema.has_index(name):
                if schema[name] == Geometry:
                    # For the geospatial footprint we need to use a special GIST index
                    result.append("CREATE INDEX idx_%s_%s ON %s USING GIST (%s);" %
                                  (self._core_table_name, name, self._core_table_name, name))
                else:
                    result.append("CREATE INDEX idx_%s_%s ON %s (%s);" %
                                  (self._core_table_name, name, self._core_table_name, name))

        # Create the tables for all non-core namespaces.
        for namespace in self._namespace_schemas:
            if namespace == "core":
                continue

            result.append(self._sql_builder.build_create_table_query(namespace))

            result.append("ALTER TABLE %s ADD COLUMN uuid UUID PRIMARY KEY;" % self._table_name(namespace))
            result.append("ALTER TABLE %s ADD CONSTRAINT %s_uuid_fkey FOREIGN KEY (uuid) REFERENCES %s (uuid) ON "
                          "DELETE CASCADE;" % (self._table_name(namespace), self._table_name(namespace),
                                               self._core_table_name))

            schema = self._namespace_schema(namespace)
            for name in schema:
                if schema.has_index(name):
                    if schema[name] == Geometry:
                        result.append("CREATE INDEX idx_%s_%s ON %s USING GIST (%s);" %
                                      (self._table_name(namespace), name, self._table_name(namespace), name))
                    else:
                        result.append("CREATE INDEX idx_%s_%s ON %s (%s);" %
                                      (self._table_name(namespace), name, self._table_name(namespace), name))

        # We use explicit 'id' primary keys for the links and tags tables so the entries can be managed using
        # other front-ends that may not support tuples as primary keys.

        # Create the table for links.
        result.append("CREATE TABLE %s (id SERIAL PRIMARY KEY, uuid UUID NOT NULL, source_uuid UUID NOT NULL);" %
                      self._link_table_name)
        result.append("ALTER TABLE %s ADD CONSTRAINT %s_link_uuid_source_uuid_uniq UNIQUE (uuid, source_uuid);" %
                      (self._link_table_name, self._link_table_name))
        result.append("ALTER TABLE %s ADD CONSTRAINT %s_uuid_fkey FOREIGN KEY (uuid) REFERENCES %s (uuid) ON "
                      "DELETE CASCADE;" % (self._link_table_name, self._link_table_name, self._core_table_name))
        result.append("CREATE INDEX idx_%s_uuid ON %s (uuid);" % (self._link_table_name, self._link_table_name))
        result.append("CREATE INDEX idx_%s_source_uuid ON %s (source_uuid);" %
                      (self._link_table_name, self._link_table_name))

        # Create the table for tags.
        result.append("CREATE TABLE %s (id SERIAL PRIMARY KEY, uuid UUID NOT NULL, tag TEXT NOT NULL);" %
                      self._tag_table_name)
        result.append("ALTER TABLE %s ADD CONSTRAINT %s_tag_uuid_tag_uniq UNIQUE (uuid, tag);" %
                      (self._tag_table_name, self._tag_table_name))
        result.append("ALTER TABLE %s ADD CONSTRAINT %s_uuid_fkey FOREIGN KEY (uuid) REFERENCES %s (uuid) ON "
                      "DELETE CASCADE;" % (self._tag_table_name, self._tag_table_name, self._core_table_name))
        result.append("CREATE INDEX idx_%s_uuid ON %s (uuid);" % (self._tag_table_name, self._tag_table_name))
        result.append("CREATE INDEX idx_%s_tag ON %s (tag);" % (self._tag_table_name, self._tag_table_name))
        return result

    def _delete_product_properties(self, uuid):
        cursor = self._connection.cursor()
        try:
            cursor.execute("DELETE FROM %s WHERE source_uuid = %s" % (self._link_table_name, self._placeholder()),
                           (uuid,))
            cursor.execute("DELETE FROM %s WHERE uuid = %s" % (self._core_table_name, self._placeholder()), (uuid,))
            assert cursor.rowcount <= 1

            if cursor.rowcount != 1:
                raise Error("could not delete properties for product: %s" % (uuid,))
        finally:
            cursor.close()

    def _derived_products(self, uuid):
        query = "SELECT uuid FROM %s WHERE source_uuid = %s" % (self._link_table_name, self._placeholder())
        parameters = (uuid,)

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
            return [row[0] for row in cursor]
        finally:
            cursor.close()

    def _drop_tables(self):
        with self._connection:
            cursor = self._connection.cursor()
            try:
                cursor.execute("DROP TABLE IF EXISTS %s CASCADE" % self._tag_table_name)
                cursor.execute("DROP TABLE IF EXISTS %s CASCADE" % self._link_table_name)
                for namespace in self._namespace_schemas:
                    if namespace != "core":
                        cursor.execute("DROP TABLE IF EXISTS %s CASCADE" % self._table_name(namespace))
                # remove 'core' table last
                if "core" in self._namespace_schemas:
                    cursor.execute("DROP TABLE IF EXISTS %s CASCADE" % self._core_table_name)
            finally:
                cursor.close()

    def _execute_list(self, sql_list):
        cursor = self._connection.cursor()
        try:
            for item in sql_list:
                cursor.execute(item)
        finally:
            cursor.close()

    def _find_products_without_available_source(self, product_type=None, grace_period=datetime.timedelta()):
        core_properties = list(self._namespace_schema("core"))
        select_list = ["%s.%s" % (self._core_table_name, name) for name in core_properties]

        query = "SELECT %s FROM %s WHERE active AND now() AT TIME ZONE 'UTC' - archive_date > %s AND uuid IN (SELECT " \
                "uuid FROM %s EXCEPT SELECT DISTINCT link.uuid FROM %s AS link LEFT JOIN %s AS source ON " \
                "(link.source_uuid = source.uuid) WHERE source.uuid IS NULL OR source.archive_path IS NOT NULL)" % \
                (", ".join(select_list), self._core_table_name, self._placeholder(), self._link_table_name,
                 self._link_table_name, self._core_table_name)

        if product_type is not None:
            query = "%s AND product_type = %s" % (query, self._placeholder())

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, (grace_period,) if product_type is None else (grace_period, product_type))

            description = [("core", core_properties)]
            return [self._unpack_product_properties(description, row) for row in cursor]
        finally:
            cursor.close()

    def _find_products_without_source(self, product_type=None, grace_period=datetime.timedelta(),
                                      archived_only=False):
        core_properties = list(self._namespace_schema("core"))
        select_list = ["%s.%s" % (self._core_table_name, name) for name in core_properties]
        query = "SELECT %s FROM %s WHERE %s.active AND now() AT TIME ZONE 'UTC' - %s.archive_date > %s AND NOT " \
                "EXISTS (SELECT 1 FROM %s WHERE %s.uuid = %s.uuid)" % (", ".join(select_list),
                                                                       self._core_table_name, self._core_table_name,
                                                                       self._core_table_name, self._placeholder(),
                                                                       self._link_table_name, self._link_table_name,
                                                                       self._core_table_name)

        if product_type is not None:
            query = "%s AND product_type = %s" % (query, self._placeholder())

        if archived_only:
            query = "%s AND archive_path IS NOT NULL" % query

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, (grace_period,) if product_type is None else (grace_period, product_type))

            description = [("core", core_properties)]
            return [self._unpack_product_properties(description, row) for row in cursor]
        finally:
            cursor.close()

    def _insert_namespace_properties(self, uuid, name, properties):
        self._validate_namespace_properties(name, properties)
        assert uuid is not None and getattr(properties, "uuid", uuid) == uuid

        # Split the properties into a list of (database) field names and a list of values. This assumes the database
        # field that corresponds to a given property has the same name. If the backend uses different field names, the
        # required translation can be performed here. Values can also be translated if necessary.
        properties_dict = vars(properties)
        fields, parameters = dictkeys(properties_dict), dictvalues(properties_dict)
        # Ensure the uuid field is present (for namespaces other than the core namespace this is used as the foreign
        # key).
        if "uuid" not in properties:
            fields.append("uuid")
            parameters.append(uuid)

        # TODO generic conversion? more overlap between sqlite.py and postgresql.py?
        parameters = [json.dumps(p) if isinstance(p, dict) else p for p in parameters]

        # Build and execute INSERT query.
        query = "INSERT INTO %s (%s) VALUES (%s)" % (self._table_name(name), ", ".join(fields),
                                                     ", ".join([self._placeholder()] * len(fields)))

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
        finally:
            cursor.close()

    def _delete_namespace_properties(self, uuid, name):
        query = "DELETE FROM %s WHERE uuid=%s" % (self._table_name(name), self._placeholder())
        cursor = self._connection.cursor()
        try:
            cursor.execute(query, [uuid])
        finally:
            cursor.close()

    def _swallow_unique_violation(self, _error):
        # There is still a small chance due to concurrency that a link/tag already exists.
        # For those cases we swallow the exception.
        swallow = False

        if self._library == 'psycopg2':
            try:
                if _error.pgcode == PG_UNIQUE_VIOLATION:
                    swallow = True
            except AttributeError:
                pass

        elif self._library == 'pg8000':  # TODO positional - issue filed on github
            try:
                if _error.args[0]['C'] == PG_UNIQUE_VIOLATION:
                    swallow = True
            except (TypeError, IndexError, AttributeError, KeyError):
                pass

        return swallow

    def _link(self, uuid, source_uuids):
        query = "INSERT INTO %s (uuid, source_uuid) SELECT %s, %s" % (self._link_table_name, self._placeholder(),
                                                                      self._placeholder())

        query += " WHERE NOT EXISTS (SELECT 1 FROM %s WHERE uuid=%s and source_uuid=%s)" % (self._link_table_name,
                                                                                            self._placeholder(),
                                                                                            self._placeholder())
        for source_uuid in source_uuids:
            with self._connection:
                cursor = self._connection.cursor()
                try:
                    cursor.execute(query, (uuid, source_uuid, uuid, source_uuid))
                except self._connection._backend.Error as _error:
                    if not self._swallow_unique_violation(_error):
                        raise
                finally:
                    cursor.close()

    def _namespace_schema(self, namespace):
        try:
            return self._namespace_schemas[namespace]
        except KeyError:
            raise Error("undefined namespace: \"%s\"" % namespace)

    def _placeholder(self, name=None, arg=None):
        if name is not None:
            result = "%%(%s)s" % name
            if isinstance(arg, datetime.datetime):
                result += "::timestamp"
        else:
            result = "%s"
        return result

    def _rewriter_property(self, column_name, subscript):
        # timestamp
        if subscript == 'year':
            return "TO_CHAR(%s, 'YYYY')" % column_name
        if subscript == 'month':
            return "TO_CHAR(%s, 'MM')" % column_name
        if subscript == 'yearmonth':
            return "TO_CHAR(%s, 'YYYY-MM')" % column_name
        if subscript == 'day':
            return "TO_CHAR(%s, 'DD')" % column_name
        if subscript == 'date':
            return "TO_CHAR(%s, 'YYYY-MM-DD')" % column_name
        if subscript == 'hour':
            return "TO_CHAR(%s, 'HH24')" % column_name
        if subscript == 'minute':
            return "TO_CHAR(%s, 'MI')" % column_name
        if subscript == 'second':
            return "TO_CHAR(%s, 'SS')" % column_name
        if subscript == 'time':
            return "TO_CHAR(%s, 'HH24:MI:SS')" % column_name
        # text
        if subscript == 'length':
            return "CHAR_LENGTH(%s)" % column_name
        raise ValueError('Unsupported subscript: %s' % subscript)

    def _rewriter_table(self):
        rewriter_table = sql.default_rewriter_table()

        #
        # Timestamp binary minus operator.
        #
        rewriter_table[Prototype("-", (Timestamp, Timestamp), Real)] = \
            lambda arg0, arg1: "EXTRACT(EPOCH FROM (%s) - (%s))" % (arg0, arg1)

        #
        # Functions.
        #
        rewriter_table[Prototype("covers", (Geometry, Geometry), Boolean)] = \
            sql.binary_function_rewriter("ST_Covers")

        # use faster non-spheroid distance calculation by passing use_spheroid = false
        rewriter_table[Prototype("distance", (Geometry, Geometry), Real)] = \
            lambda arg0, arg1: "ST_Distance(%s, %s, false)" % (arg0, arg1)

        rewriter_table[Prototype("intersects", (Geometry, Geometry), Boolean)] = \
            sql.binary_function_rewriter("ST_Intersects")

        rewriter_table[Prototype("is_source_of", (UUID,), Boolean)] = \
            lambda arg0: "EXISTS (SELECT 1 FROM %s WHERE source_uuid = %s.uuid AND uuid = (%s))" % \
            (self._link_table_name, self._core_table_name, arg0)

        def is_source_of_subquery(where_expr, where_namespaces):
            joins = ''
            for namespace in where_namespaces:
                joins = "%s INNER JOIN %s USING (uuid)" % (joins, self._table_name(namespace))

            return "{core}.uuid in (SELECT {link}.source_uuid FROM {core} {joins} " \
                "INNER JOIN {link} on {link}.uuid = {core}.uuid WHERE {where})".format(
                    core=self._core_table_name, link=self._link_table_name, joins=joins,
                    where=where_expr)

        rewriter_table[Prototype("is_source_of", (Boolean,), Boolean)] = is_source_of_subquery

        rewriter_table[Prototype("is_derived_from", (UUID,), Boolean)] = \
            lambda arg0: "EXISTS (SELECT 1 FROM %s WHERE uuid = %s.uuid AND source_uuid = (%s))" % \
            (self._link_table_name, self._core_table_name, arg0)

        def is_derived_from_subquery(where_expr, where_namespaces):
            joins = ''
            for namespace in where_namespaces:
                joins = "%s INNER JOIN %s USING (uuid)" % (joins, self._table_name(namespace))

            return "{core}.uuid in (SELECT {link}.uuid FROM {core} {joins} " \
                "INNER JOIN {link} on {link}.source_uuid = {core}.uuid WHERE {where})".format(
                    core=self._core_table_name, link=self._link_table_name, joins=joins,
                    where=where_expr)

        rewriter_table[Prototype("is_derived_from", (Boolean,), Boolean)] = is_derived_from_subquery

        rewriter_table[Prototype("has_tag", (Text,), Boolean)] = \
            lambda arg0: "EXISTS (SELECT 1 FROM %s WHERE uuid = %s.uuid AND tag = (%s))" % \
            (self._tag_table_name, self._core_table_name, arg0)

        rewriter_table[Prototype("now", (), Timestamp)] = \
            sql.as_is("now() AT TIME ZONE 'UTC'")

        def is_defined_rewriter(arg):
            namespace_name = arg.split('.')
            if len(namespace_name) == 1:
                return 'EXISTS (SELECT 1 FROM %s WHERE uuid = %s.uuid)' % \
                    (self._table_name(arg), self._core_table_name)
            else:
                return "(%s) IS NOT NULL" % arg

        rewriter_table[Prototype("is_defined", (Long,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Integer,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Real,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Boolean,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Text,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Namespace,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Timestamp,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (UUID,), Boolean)] = is_defined_rewriter
        rewriter_table[Prototype("is_defined", (Geometry,), Boolean)] = is_defined_rewriter

        return rewriter_table

    def _source_products(self, uuid):
        query = "SELECT source_uuid FROM %s WHERE uuid = %s" % (self._link_table_name, self._placeholder())
        parameters = (uuid,)

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
            return [row[0] for row in cursor]
        finally:
            cursor.close()

    def _table_name(self, name):
        return name if not self._table_prefix else self._table_prefix + name

    def _tag(self, uuid, tags):
        query = "INSERT INTO %s (uuid, tag) SELECT %s, %s" % (self._tag_table_name, self._placeholder(),
                                                              self._placeholder())
        query += " WHERE NOT EXISTS (SELECT 1 FROM %s WHERE uuid=%s and tag=%s)" % (self._tag_table_name,
                                                                                    self._placeholder(),
                                                                                    self._placeholder())
        for tag in tags:
            with self._connection:
                cursor = self._connection.cursor()
                try:
                    cursor.execute(query, (uuid, tag, uuid, tag))
                except self._connection._backend.Error as _error:
                    if not self._swallow_unique_violation(_error):
                        raise
                finally:
                    cursor.close()

    def _tags(self, uuid):
        query = "SELECT tag FROM %s WHERE uuid = %s ORDER BY tag" % (self._tag_table_name, self._placeholder())
        parameters = (uuid,)

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
            return [row[0] for row in cursor]
        finally:
            cursor.close()

    def _type_map(self):
        type_map = sql.TypeMap()
        type_map[Long] = "BIGINT"
        type_map[Integer] = "INTEGER"
        type_map[Real] = "DOUBLE PRECISION"
        type_map[Boolean] = "BOOLEAN"
        type_map[Text] = "TEXT COLLATE \"C\""
        type_map[JSON] = "JSONB"
        type_map[Timestamp] = "TIMESTAMP"
        type_map[UUID] = "UUID"
        type_map[Geometry] = "GEOGRAPHY"

        return type_map

    def _unlink(self, uuid, source_uuids=None):
        if source_uuids is None:
            query = "DELETE FROM %s WHERE uuid = %s" % (self._link_table_name, self._placeholder())
            parameters = (uuid,)
        else:
            for source_uuid in source_uuids:
                query = "DELETE FROM %s WHERE uuid = %s AND source_uuid = %s" % \
                    (self._link_table_name, self._placeholder(), self._placeholder())
                parameters = (uuid, source_uuid)
                cursor = self._connection.cursor()
                try:
                    cursor.execute(query, parameters)
                finally:
                    cursor.close()

    def _untag(self, uuid, tags=None):
        if tags is None:
            query = "DELETE FROM %s WHERE uuid = %s" % (self._tag_table_name, self._placeholder())
            parameters = (uuid,)
            cursor = self._connection.cursor()
            try:
                cursor.execute(query, parameters)
            finally:
                cursor.close()
        else:
            for tag in tags:
                query = "DELETE FROM %s WHERE uuid = %s AND tag = %s" % (self._tag_table_name,
                                                                         self._placeholder(), self._placeholder())
                parameters = (uuid, tag)
                cursor = self._connection.cursor()
                try:
                    cursor.execute(query, parameters)
                finally:
                    cursor.close()

    def _update_namespace_properties(self, uuid, name, properties):
        self._validate_namespace_properties(name, properties, partial=True)
        assert uuid is not None and getattr(properties, "uuid", uuid) == uuid

        # Split the properties into a list of (database) field names and a list of values. This assumes the database
        # field that corresponds to a given property has the same name. If the backend uses different field names, the
        # required translation can be performed here. Values can also be translated if necessary.
        properties_dict = vars(properties)
        fields, parameters = dictkeys(properties_dict), dictvalues(properties_dict)
        if not fields:
            return  # nothing to do

        # Remove the uuid field if present. This field needs to be included in the WHERE clause of the UPDATE query,
        # not in the SET clause.
        try:
            uuid_index = fields.index("uuid")
        except ValueError:
            pass
        else:
            del fields[uuid_index]
            del parameters[uuid_index]

        # Append the uuid (value) at the end of the list of parameters (will be used in the WHERE clause).
        parameters.append(uuid)

        parameters = [json.dumps(p) if isinstance(p, dict) else p for p in parameters]

        # Build and execute UPDATE query.
        set_clause = ", ".join(["%s = %s" % (field, self._placeholder()) for field in fields])
        query = "UPDATE %s SET %s WHERE uuid = %s" % (self._table_name(name), set_clause, self._placeholder())

        cursor = self._connection.cursor()
        try:
            cursor.execute(query, parameters)
            assert cursor.rowcount <= 1

            if cursor.rowcount != 1:
                raise Error("could not update properties for namespace: %s for product: %s" % (name, uuid))
        finally:
            cursor.close()

    def _unpack_namespace_properties(self, namespace, description, values):
        unpacked_properties = Struct()
        schema = self._namespace_schema(namespace)
        for identifier, value in zip(description, values):
            # We may get unicode from the (psycopg2) connection
            # if, possibly by a third party, the UNICODE adapter is loaded.
            # Muninn assumes strs
            if is_python2_unicode(value):
                value = value.encode(self._connection.encoding)

            if value is not None or not schema.is_optional(identifier):
                unpacked_properties[identifier] = value

        return unpacked_properties

    def _unpack_product_properties(self, description, values):
        unpacked_properties, start = Struct(), 0
        for ns_name, ns_description in description:
            end = start + len(ns_description)

            # A value of None for the uuid field of namespaces other than the core namespace indicates the namespace is
            # not defined for the product under consideration. In this case, the entire namespace will be skipped.
            # Otherwise, only the uuid field is skipped, as it is an implementation detail (foreign key) and it is not
            # part of the namespace itself.
            #
            if ns_name != "core":
                assert ns_description[0] == "uuid"
                if values[start] is None:
                    # Skip the entire namespace.
                    start = end
                    continue

                # Skip the uuid field.
                start += 1
                ns_description = ns_description[1:]

            unpacked_ns_properties = self._unpack_namespace_properties(ns_name, ns_description, values[start:end])
            self._validate_namespace_properties(ns_name, unpacked_ns_properties, partial=True)
            unpacked_properties[ns_name] = unpacked_ns_properties
            start = end

        return unpacked_properties

    def _validate_namespace_properties(self, namespace, properties, partial=False):
        self._namespace_schema(namespace).validate(properties, partial)

    @translate_errors
    def count(self, where="", parameters={}):
        query, query_parameters = self._sql_builder.build_count_query(where, parameters)

        with self._connection:
            cursor = self._connection.cursor()
            try:
                cursor.execute(query, query_parameters)
                assert cursor.rowcount == 1
                return cursor.fetchone()[0]
            finally:
                cursor.close()

    @translate_errors
    def delete_product_properties(self, uuid):
        with self._connection:
            self._delete_product_properties(uuid)

    @translate_errors
    def derived_products(self, uuid):
        with self._connection:
            return self._derived_products(uuid)

    @translate_errors
    def destroy(self):
        self._drop_tables()

    @translate_errors
    def disconnect(self):
        """Drop the connection to the database in order to free up resources. The connection will be re-established
        automatically when required."""
        self._connection.close()

    @translate_errors
    def exists(self):
        with self._connection:
            query = "SELECT relname FROM pg_class WHERE relname=%s" % (self._placeholder(),)
            parameters = (self._core_table_name,)
            cursor = self._connection.cursor()
            try:
                cursor.execute(query, parameters)
                return len(cursor.fetchall()) != 0
            finally:
                cursor.close()

    @translate_errors
    def find_products_without_available_source(self, product_type=None, grace_period=datetime.timedelta()):
        """Return the core properties of all products that are linked to one or more source products, all of which are
           unavailable. A product is unavailable if there is no data associated with it, only properties. Products that
           have links to external source products will not be selected by this function, because it cannot be
           determined whether or not these products are available.

           Keyword arguments:
           product_type --  Only consider products of the specified product type.

        """
        with self._connection:
            return self._find_products_without_available_source(product_type, grace_period)

    @translate_errors
    def find_products_without_source(self, product_type=None, grace_period=datetime.timedelta(),
                                     archived_only=False):
        """Return the core properties of all products that are not linked to any source products.

           Keyword arguments:
           product_type --  Only consider products of the specified product type.

        """
        with self._connection:
            return self._find_products_without_source(product_type, grace_period, archived_only)

    def initialize(self, namespace_schemas):
        self._namespace_schemas = namespace_schemas
        self._sql_builder = sql.SQLBuilder(self._namespace_schemas, self._type_map(), self._rewriter_table(),
                                           self._table_name, self._placeholder, self._placeholder,
                                           self._rewriter_property)

    @translate_errors
    def insert_product_properties(self, properties):
        with self._connection:
            self._insert_namespace_properties(properties.core.uuid, "core", properties.core)
            for ns_name, ns_properties in vars(properties).items():
                if ns_name == "core" or ns_properties is None:
                    continue
                self._insert_namespace_properties(properties.core.uuid, ns_name, ns_properties)

    @translate_errors
    def link(self, uuid, source_uuids):
        self._link(uuid, source_uuids)

    @translate_errors
    def prepare(self, dry_run=False):
        sqls = self._create_tables_sql()
        if not dry_run:
            with self._connection:
                self._execute_list(sqls)
        return sqls

    @translate_errors
    def search(self, where="", order_by=[], limit=None, parameters={}, namespaces=[], property_names=[]):
        query, query_parameters, query_description = \
            self._sql_builder.build_search_query(where, order_by, limit, parameters, namespaces, property_names)

        with self._connection:
            cursor = self._connection.cursor()
            try:
                cursor.execute(query, query_parameters)
                return [self._unpack_product_properties(query_description, row) for row in cursor]
            finally:
                cursor.close()

    @translate_errors
    def server_time_utc(self):
        with self._connection:
            cursor = self._connection.cursor()
            try:
                cursor.execute("SELECT timezone(%s, now())", ("UTC",))
                assert cursor.rowcount == 1
                return cursor.fetchone()[0]
            finally:
                cursor.close()

    @translate_errors
    def source_products(self, uuid):
        with self._connection:
            return self._source_products(uuid)

    @translate_errors
    def summary(self, where="", parameters=None, aggregates=None, group_by=None, group_by_tag=False,
                having=None, order_by=None):
        query, query_parameters, query_description = self._sql_builder.build_summary_query(
            where, parameters, aggregates, group_by, group_by_tag, having, order_by)

        with self._connection:
            cursor = self._connection.cursor()
            try:
                cursor.execute(query, query_parameters)
                return [row for row in cursor], query_description
            finally:
                cursor.close()

    @translate_errors
    def tag(self, uuid, tags):
        self._tag(uuid, tags)

    @translate_errors
    def tags(self, uuid):
        with self._connection:
            return self._tags(uuid)

    @translate_errors
    def unlink(self, uuid, source_uuids=None):
        with self._connection:
            self._unlink(uuid, source_uuids)

    @translate_errors
    def untag(self, uuid, tags=None):
        with self._connection:
            self._untag(uuid, tags)

    @translate_errors
    def update_product_properties(self, properties, uuid=None, new_namespaces=None):
        new_namespaces = new_namespaces or []
        if "core" in properties:
            self._validate_namespace_properties("core", properties.core, partial=True)
            if "uuid" in properties.core:
                uuid = properties.core.uuid if uuid is None else uuid
                if uuid != properties.core.uuid:
                    raise Error("specified uuid does not match uuid included in the specified product properties")

        if uuid is None:
            raise Error("no uuid specified and no uuid included in the specified product properties")

        with self._connection:
            self._update_namespace_properties(uuid, "core", properties.core)
            for ns_name, ns_properties in vars(properties).items():
                if ns_name == "core":
                    continue
                if ns_name in new_namespaces:
                    if ns_properties is not None:
                        self._insert_namespace_properties(uuid, ns_name, ns_properties)
                elif ns_properties is None:
                    self._delete_namespace_properties(uuid, ns_name)
                else:
                    self._update_namespace_properties(uuid, ns_name, ns_properties)



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import re
from datetime import timedelta
import argparse

import muninn
from muninn.tools.utils import create_parser, parse_args_and_run


try:
    import tabulate
except ImportError:
    tabulate = None

OWN_OUTPUT_FORMATS = ['psv', 'csv']
if tabulate is None:
    DEFAULT_FORMAT = 'psv'
    OUTPUT_FORMATS = OWN_OUTPUT_FORMATS
else:
    DEFAULT_FORMAT = 'orgtbl'
    OUTPUT_FORMATS = set(tabulate.tabulate_formats + OWN_OUTPUT_FORMATS)


DEFAULT_STATS = ['size.sum', 'validity_start.min', 'validity_stop.max']


def ceil(size):
    integer_size = int(size)
    return integer_size + 1 if size > integer_size else integer_size


def human_readable_size(size, base=1024, powers=["Bi", "Ki", "Mi", "Gi", "Ti", "Ei"]):
    if len(powers) == 0:
        return str(size)

    power = 0
    unit = 1
    while power < len(powers) - 1 and unit * base <= size:
        unit *= base
        power += 1

    size = size / unit
    ceil_size = ceil(size)
    ceil_size_10 = ceil(size * 10.0) / 10.0

    if power > 0 and size < 10.0 and ceil_size_10 < 10.0:
        result = "%.1f" % ceil_size_10
    elif ceil_size == base and power < len(powers) - 1:
        power += 1
        result = "1.0"
    else:
        result = str(ceil_size)

    return result + powers[power]


def format_duration(duration):
    if duration is None:
        return ''
        # return "<unknown>"

    try:
        duration = timedelta(seconds=round(duration))
    except OverflowError:
        return "<overflow>"
    return str(duration)


def format_size(size, human_readable=False):
    if size is None:
        return ''
        # return "<unknown>"

    if not human_readable:
        return str(size)

    return human_readable_size(float(size))


# Support multiple table output formats

class PlainWriter(object):
    def __init__(self, header, human_readable=False):
        self._header = [re.sub('^core.', '', item) for item in header]
        self._duration_fields = []
        self._size_fields = []
        self.human_readable = human_readable
        for i, name in enumerate(header):
            if name.startswith('core.validity_duration'):
                self._duration_fields.append(i)
            elif name.startswith('core.size'):
                self._size_fields.append(i)

    def header(self):
        print("|", " | ".join(self._header), "|")

    def _format_items(self, values):
        result = []
        for i, value in enumerate(values):
            if i in self._duration_fields:
                result.append(format_duration(value))
            elif i in self._size_fields:
                result.append(format_size(value, self.human_readable))
            else:
                result.append(value)
        return result

    def row(self, values):
        print("|", " | ".join(str(item) for item in self._format_items(values)), "|")

    def footer(self):
        pass


class TabulateWriter(PlainWriter):
    def __init__(self, header, human_readable=False, fmt='orgtbl'):
        super(TabulateWriter, self).__init__(header, human_readable)
        self._data = []
        self._format = fmt

    def header(self):
        pass

    def row(self, values):
        self._data.append(self._format_items(values))

    def footer(self):
        # align core.size fields to the right (only available in tabulate > 0.8.2)
        # colalign = ['right' if i in self._size_fields else None for i in range(len(self._header))]
        # print(tabulate.tabulate(self._data, headers=self._header, tablefmt=self._format), colalign=colalign)
        print(tabulate.tabulate(self._data, headers=self._header, tablefmt=self._format))


class CSVWriter(PlainWriter):
    def header(self):
        print(",".join(["\"" + name.replace("\"", "\"\"") + "\"" for name in self._header]))

    def row(self, values):
        print(",".join("\"" + str(item).replace("\"", "\"\"") + "\"" for item in self._format_items(values)))


def get_writer(header, args):
    if args.output_format == "psv":  # PSV = Pipe Separated Values
        writer = PlainWriter(header, args.human_readable)
    elif args.output_format == "csv":
        writer = CSVWriter(header, args.human_readable)
    elif tabulate is not None:
        writer = TabulateWriter(header, args.human_readable, args.output_format)
    else:
        writer = PlainWriter(header, args.human_readable)
    return writer


def canonical_property(archive, name):
    '''
    Make sure "core" is added to properties without namespace.
    '''
    if name in ['count', 'tag']:
        return name

    metadata = archive._namespace_schemas

    if name == '*':
        name = '*.*'
    else:
        split_name = name.split('.')
        if split_name[0] not in metadata and (split_name[0] in metadata['core'] or
                                              split_name[0] == 'validity_duration'):
            name = '%s.%s' % ('core', name)

    return name


def coalesce_identifier_arguments(arg, archive):
    if arg:
        names = re.split('[ ,]+', ' '.join(arg))
        result = [canonical_property(archive, name) for name in names if name]
    else:
        result = []
    return result


def coalesce_order_by_args(arg, archive):
    # An order specifier without a "+" (ascending) prefix is interpreted as descending. Otherwise, it would be
    # impossible to specify descending order on the command line, because a "-" prefix is interpreted as an option by
    result = []
    if arg:
        names = re.split('[ ,]+', ' '.join(arg))
        for name in names:
            if name:
                if name.startswith('+') or name.startswith('-'):
                    order = name[0]
                    name = name[1:]
                else:
                    order = '-'
                name = canonical_property(archive, name)
                result.append(order + name)
    return result


def run(args):
    with muninn.open(args.archive) as archive:
        group_by = coalesce_identifier_arguments(args.group_by, archive)
        if args.stats:
            stats = coalesce_identifier_arguments(args.stats, archive)
        else:
            stats = coalesce_identifier_arguments(DEFAULT_STATS, archive)
        order_by = coalesce_order_by_args(args.order_by, archive)

        result, header = archive.summary(
            args.expression,
            aggregates=stats,
            group_by=group_by,
            group_by_tag=args.group_by_tag,
            order_by=order_by,
            having=args.stats_filter
        )

        # Output summary in the requested output format.
        writer = get_writer(header, args)
        writer.header()
        for product in result:
            writer.row(product)
        writer.footer()

    return 0


def main():
    parser = create_parser(
        description="Summary of the products matching the search expression.",
        formatter_class=argparse.RawDescriptionHelpFormatter)
    parser.add_argument("-f", "--output-format", choices=OUTPUT_FORMATS, default=DEFAULT_FORMAT,
                        metavar="FORMAT", help="output format; choices (depending on installation of "
                        "python-tabulate): {%(choices)s}; default: %(default)s")
    parser.add_argument("-o", "--order-by", action="append", dest="order_by", help="white space "
                        "separated list of sort order specifiers; a \"+\" prefix denotes ascending order; no prefix "
                        "denotes descending order")
    parser.add_argument("-g", "--group-by", action="append", dest="group_by",
                        help="list of property names whose values are used for grouping the aggregation results; "
                        "there will be a separate result row for each combination of group_by property values; "
                        "properties need to be of type long, integer, boolean, text or timestamp; "
                        "timestamps require a binning subscript which can be 'year', 'month', 'yearmonth', 'day', "
                        "'date', 'hour', 'minute', 'second' or 'time' (e.g. `validity_start.yearmonth`)")
    parser.add_argument("-t", "--group-by-tag", action="store_true", help="group result by tag value; "
                        "note that products will be counted multiple times if they have multiple tags")
    parser.add_argument("-s", "--stats", action="append", dest="stats", help="list of property aggregates defined as "
                        "`<property_name>.<reduce_fn>`; "
                        "properties need to be of type long, integer, real, text or timestamp; "
                        "the reduce function can be 'min', 'max', 'sum', or 'avg'; "
                        "'sum' and 'avg' are not possible for text and timestamp properties; "
                        "a special property 'validity_duration' (defined as validity_stop - validity_start) can also "
                        "be used; default stats: %r" % ' '.join(DEFAULT_STATS))
    parser.add_argument("--stats-filter", metavar='EXPRESSION', help="expression used to filter on aggregates")
    parser.add_argument("-H", "--human-readable", action="store_true", help="output human readable core.size")
    parser.add_argument("archive", metavar="ARCHIVE", help="identifier of the archive to use")
    parser.add_argument("expression", metavar="EXPRESSION", nargs='?', help="expression used to search for products")

    return parse_args_and_run(parser, run)


if __name__ == '__main__':
    main()



import json
import logging
import os

from .base import StorageBackend

from muninn.schema import Mapping, Text, Integer
import muninn.config as config
import muninn.util as util
from muninn.exceptions import Error, StorageError

import boto3
import boto3.s3
import botocore

logging.getLogger("boto3").setLevel(logging.CRITICAL)


class _S3Config(Mapping):
    _alias = "s3"

    host = Text()
    port = Integer(optional=True)
    bucket = Text()
    access_key = Text(optional=True)
    secret_access_key = Text(optional=True)
    region = Text(optional=True)
    prefix = Text(optional=True)
    download_args = Text(optional=True)  # JSON representation of boto3 download_file ExtraArgs parameter
    upload_args = Text(optional=True)  # JSON representation of boto3 upload_file ExtraArgs parameter
    copy_args = Text(optional=True)  # JSON representation of boto3 copy ExtraArgs parameter
    transfer_config = Text(optional=True)  # JSON representation of boto3.s3.transfer.TransferConfig parameters


def create(configuration, tempdir, auth_file):
    options = config.parse(configuration.get("s3", {}), _S3Config)

    # if access_key and secret_access_key missing, use auth_file
    if (auth_file is not None and
            'access_key' not in options and
            'secret_access_key' not in options and
            'host' in options and
            'bucket' in options):
        credentials = json.loads(open(auth_file).read())
        s3url = "s3://" + options['bucket']
        if options['host'] in credentials:
            record = credentials[options['host']]
            if record.get('auth_type') == 'S3' and record.get('bucket') == options['bucket']:
                for option in ('access_key', 'secret_access_key', 'port', 'region'):
                    if option in record and option not in options:
                        options[option] = record[option]
        elif s3url in credentials:
            record = credentials[s3url]
            if record.get('host') == options['host']:
                for option in ('access_key', 'secret_access_key', 'port', 'region'):
                    if option in record and option not in options:
                        options[option] = record[option]

    # check that mandatory options are configured
    for option in ('access_key', 'secret_access_key'):
        if option not in options:
            raise Error("'%s' not configured" % option)

    _S3Config.validate(options)
    return S3StorageBackend(tempdir=tempdir, **options)


class S3StorageBackend(StorageBackend):  # TODO '/' in keys to indicate directory, 'dir/' with contents?
    def __init__(self, bucket, host, access_key, secret_access_key, port=None, region=None, prefix='',
                 download_args=None, upload_args=None, copy_args=None, transfer_config=None, tempdir=None):
        super(S3StorageBackend, self).__init__(tempdir)

        self.bucket = bucket

        if prefix and not prefix.endswith('/'):
            prefix += '/'
        self._prefix = prefix

        endpoint_url = host
        if ':' not in host:
            if port == 443:
                endpoint_url = 'https://' + endpoint_url
            else:
                endpoint_url = 'http://' + endpoint_url
                if port is not None and port != 80:
                    endpoint_url += ':%d' % port
        elif port is not None:
            endpoint_url += ':%d' % port
        self.global_prefix = util.fwd_join(endpoint_url, bucket, prefix)

        self._root = bucket

        self._resource = boto3.resource(
            service_name='s3',
            region_name=region,
            aws_access_key_id=access_key,
            aws_secret_access_key=secret_access_key,
            endpoint_url=endpoint_url,
        )

        self._download_args = None
        if download_args:
            self._download_args = json.loads(download_args)
        self._upload_args = None
        if upload_args:
            self._upload_args = json.loads(upload_args)
        self._copy_args = None
        if copy_args:
            self._copy_args = json.loads(copy_args)
        if transfer_config:
            self._transfer_config = boto3.s3.transfer.TransferConfig(**json.loads(transfer_config))
        else:
            self._transfer_config = boto3.s3.transfer.TransferConfig()

    def _bucket_exists(self):
        try:
            self._resource.meta.client.head_bucket(Bucket=self.bucket)
            return True
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                return False
            else:
                raise

    def _prefix_exists(self):
        if self._prefix:
            objs = list(self._resource.Bucket(self.bucket).objects.limit(count=1).filter(Prefix=self._prefix))
            return len(objs) == 1
        else:
            return True

    def prepare(self):
        if not self._bucket_exists():
            self._resource.create_bucket(Bucket=self.bucket)
        if not self._prefix_exists():
            self._create_dir(self._prefix)

    def exists(self):
        return self._bucket_exists() and self._prefix_exists()

    def destroy(self):
        if self._prefix:
            self._resource.Bucket(self.bucket).objects.filter(Prefix=self._prefix).delete()
        elif self._bucket_exists():
            bucket = self._resource.Bucket(self.bucket)
            bucket.objects.all().delete()
            bucket.delete()

    def current_archive_path(self, paths, properties):
        raise Error("S3 storage backend does not support ingesting already archived products")

    def _upload_file(self, key, path):
        key = key.replace('\\', '/')
        obj = self._resource.Object(self.bucket, key)
        if os.path.getsize(path) == 0:  # upload_file can hang on empty files
            self._resource.Object(self.bucket, key).put()
        else:
            obj.upload_file(path, ExtraArgs=self._upload_args, Config=self._transfer_config)

    def _create_dir(self, key):
        # using put, as upload_file/upload_fileobj do not like the trailing slash
        key = key.replace('\\', '/')
        if not key.endswith('/'):
            key = key + '/'
        self._resource.Object(self.bucket, key).put()

    def put(self, paths, properties, use_enclosing_directory, use_symlinks=None,
            retrieve_files=None, run_for_product=None):

        if use_symlinks:
            raise Error("S3 storage backend does not support symlinks")

        anything_stored = False

        try:
            archive_path = properties.core.archive_path
            physical_name = properties.core.physical_name

            if not use_enclosing_directory and retrieve_files is None:
                assert len(paths) == 1 and os.path.basename(paths[0]) == physical_name

            tmp_root = self.get_tmp_root(properties)
            with util.TemporaryDirectory(dir=tmp_root, prefix=".put-",
                                         suffix="-%s" % properties.core.uuid.hex) as tmp_path:
                if retrieve_files:
                    paths = retrieve_files(tmp_path)

                # Upload file(s)
                for path in paths:
                    key = self._prefix + util.fwd_join(archive_path, physical_name)

                    # Add enclosing dir
                    if use_enclosing_directory:
                        key = util.fwd_join(key, os.path.basename(path))

                    if os.path.isdir(path):
                        self._create_dir(key)
                        anything_stored = True

                        for root, subdirs, files in os.walk(path):
                            rel_root = os.path.relpath(root, path)

                            for subdir in subdirs:
                                dirkey = os.path.normpath(os.path.join(key, rel_root, subdir))
                                self._create_dir(dirkey)
                                anything_stored = True

                            for filename in files:
                                filekey = os.path.normpath(os.path.join(key, rel_root, filename))
                                filepath = util.fwd_join(root, filename)
                                self._upload_file(filekey, filepath)
                                anything_stored = True

                    else:
                        self._upload_file(key, path)
                        anything_stored = True

                if run_for_product is not None:
                    run_for_product(paths)

        except Exception as e:
            raise StorageError(e, anything_stored)

    def get(self, product, target_path, use_enclosing_directory, use_symlinks=None):
        paths = set()

        if use_symlinks:
            raise Error("S3 storage backend does not support symlinks")

        archive_path = product.core.archive_path
        product_path = util.fwd_join(archive_path, product.core.physical_name)
        prefix = self._prefix + product_path

        objs = list(self._resource.Bucket(self.bucket).objects.filter(Prefix=prefix))
        if not objs:
            raise Error("no data for product '%s' (%s)" % (product.core.product_name, product.core.uuid))

        for obj in objs:
            rel_path = os.path.relpath(obj.key, self._prefix + archive_path).replace('\\', '/')
            if use_enclosing_directory:
                rel_path = '/'.join(rel_path.split('/')[1:])
            paths.add(os.path.normpath(os.path.join(target_path, rel_path.split('/')[0])))
            target = os.path.normpath(os.path.join(target_path, rel_path))

            if obj.key.endswith('/'):
                util.make_path(target)
            else:
                dirname = os.path.dirname(target)
                if dirname != '':
                    util.make_path(dirname)
                self._resource.Object(self.bucket, obj.key).download_file(target, ExtraArgs=self._download_args,
                                                                          Config=self._transfer_config)

        return list(paths)

    def delete(self, product_path, properties):
        prefix = self._prefix + product_path
        prefix = prefix.replace('\\', '/')
        for obj in self._resource.Bucket(self.bucket).objects.filter(Prefix=prefix):
            obj.delete()

    def size(self, product_path):
        total = 0
        prefix = self._prefix + product_path
        prefix = prefix.replace('\\', '/')
        for obj in self._resource.Bucket(self.bucket).objects.filter(Prefix=prefix):
            total += obj.size
        return total

    def move(self, product, archive_path, paths=None):
        # Ignore if product already there
        if product.core.archive_path == archive_path:
            return paths

        product_path = self._prefix + util.fwd_join(product.core.archive_path, product.core.physical_name)
        new_product_path = self._prefix + util.fwd_join(archive_path, product.core.physical_name)

        objs = list(self._resource.Bucket(self.bucket).objects.filter(Prefix=product_path))
        if not objs:
            raise Error("no data for product '%s' (%s)" % (product.core.product_name, product.core.uuid))

        for obj in objs:
            new_key = os.path.normpath(os.path.join(new_product_path, os.path.relpath(obj.key, product_path)))
            new_key = new_key.replace('\\', '/')
            if obj.key.endswith('/'):
                self._create_dir(new_key)
            else:
                self._resource.Object(self.bucket, new_key).copy(CopySource={'Bucket': self.bucket, 'Key': obj.key},
                                                                 ExtraArgs=self._copy_args,
                                                                 Config=self._transfer_config)
            self._resource.Object(self.bucket, obj.key).delete()

        return paths



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

__version__ = "6.4"
__copyright__ = "Copyright (C) 2014-2024 S[&]T, The Netherlands."

__all__ = ["Error", "InternalError", "Struct", "Archive", "open", "config_path"]

import os as _os
import sys as _sys

from muninn._compat import urlparse as _urlparse
from muninn.archive import Archive
from muninn.exceptions import *
from muninn.struct import Struct


def config_path():
    """Return the value of the `MUNINN_CONFIG_PATH` environment variable."""
    return _os.environ.get("MUNINN_CONFIG_PATH", "")


def open(id=None, **kwargs):
    """Open an archive for the given archive id, by searching for the
    corresponding configuration file in the locations found in the
    `MUNINN_CONFIG_PATH` environment variable.
    Alternatively, the id can be a path/url that points directly to a
    muninn archive configuration file (Python 3 only). This path/url
    should still end with a `.cfg` extension.

    Arguments:
    id  --  Archive id (name of configuration file) or path/url to a configuration file

    Returns:
    An instance of `muninn.archive.Archive`
    """
    if id is None:
        configuration = {}
    else:
        configuration = _read_archive_config_file(_locate_archive_config_file(id))
        if id.endswith(".cfg"):
            if "/" in id:
                id = id.rsplit("/", 1)[1]
            id = _os.path.splitext(id)[0]

    for name, value in kwargs.items():
        section = configuration.setdefault(name, {})
        section.update(value)

    return Archive.create(configuration, id)


def list_archives():
    """Search locations found in the 'MUNINN_CONFIG_PATH' environment variable
    and return names of all found archives."""
    result = []
    for path in filter(None, config_path().split(":")):
        if _os.path.isfile(path):
            result.append(_os.path.basename(path)[:-4])
        else:
            result.extend([f[:-4] for f in _os.listdir(path) if f.endswith('.cfg')])

    return result


def _read_archive_config_file(path):
    try:
        from ConfigParser import ConfigParser
    except ImportError:
        from configparser import ConfigParser

    parser = ConfigParser()

    if _urlparse(path).scheme != '':
        if _sys.version_info[0] < 3:
            raise NotImplementedError("retrieving archive configuration files via an url requires python 3")
        import urllib.request
        with urllib.request.urlopen(path, timeout=60) as response:
            parser.read_string(response.read().decode("utf-8"))
    else:
        if not parser.read(path):
            raise Error("unable to read config file: \"%s\"" % path)

    return dict([(name, dict(parser.items(name))) for name in parser.sections()])


def _locate_archive_config_file(archive_id):
    if archive_id.endswith(".cfg"):
        # the archive id is already a path/url to the config file
        return archive_id
    config_file_name = "%s.cfg" % archive_id
    if _os.path.basename(config_file_name) != config_file_name:
        raise Error("invalid archive identifier: \"%s\"" % archive_id)

    for path in filter(None, config_path().split(":")):
        if _os.path.isfile(path):
            if _os.path.basename(path) == config_file_name:
                return path
        else:
            config_file_path = _os.path.join(path, config_file_name)
            if _os.path.isfile(config_file_path):
                return config_file_path

    raise Error("configuration file: \"%s\" not found on search path: \"%s\"" % (config_file_name, config_path()))



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function


class Error(Exception):
    pass


class InternalError(Error):
    pass


class StorageError(Error):
    def __init__(self, orig, anything_stored):
        self.orig = orig
        self.anything_stored = anything_stored



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import muninn

from muninn.tools.utils import Processor, create_parser, parse_args_and_run


class RemoveProcessor(Processor):
    def __init__(self, args):
        super(RemoveProcessor, self).__init__(args)
        self.args = args

    def perform_operation(self, archive, product):
        if self.args.catalogue_only:
            archive.delete_properties(product.core.uuid)
        else:
            archive.remove(product.core.uuid, force=self.args.force, cascade=False)
        return 1


def remove(args):
    processor = RemoveProcessor(args)
    with muninn.open(args.archive) as archive:
        products = archive.search(where=args.expression, property_names=['uuid'])
        returncode = processor.process(archive, args, products)
        if not args.catalogue_only:
            archive.cleanup_derived_products()
        return returncode


def main():
    parser = create_parser(description="Remove products from a muninn archive.", parallel=True)
    parser.add_argument("-c", "--catalogue-only", action="store_true", help="remove the entry from the catalogue "
                        "without removing any product from the storage")
    parser.add_argument("-f", "--force", action="store_true", help="also remove partially ingested products; note "
                        "that this can cause products to be removed while in the process of being ingested")
    parser.add_argument("archive", metavar="ARCHIVE", help="identifier of the archive to use")
    parser.add_argument("expression", metavar="EXPRESSION", help="expression used to search for products to remove")
    return parse_args_and_run(parser, remove)


if __name__ == '__main__':
    main()



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import inspect


class TypeVisitor(object):
    def visit(self, visitable, *args, **kwargs):
        for type in inspect.getmro(visitable):
            try:
                visit_func = getattr(self, "visit_%s" % type.__name__)
            except AttributeError:
                pass
            else:
                return visit_func(visitable, *args, **kwargs)

        try:
            visit_func = getattr(self, "default")
        except AttributeError:
            pass
        else:
            return visit_func(visitable, *args, **kwargs)


class Visitor(object):
    def visit(self, visitable, *args, **kwargs):
        for type_ in inspect.getmro(type(visitable)):
            try:
                visit_func = getattr(self, "visit_%s" % type_.__name__)
            except AttributeError:
                pass
            else:
                return visit_func(visitable, *args, **kwargs)

        try:
            visit_func = getattr(self, "default")
        except AttributeError:
            pass
        else:
            return visit_func(visitable, *args, **kwargs)



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

from muninn._compat import string_types as basestring

import copy
import datetime
import re
import uuid

import muninn.geometry as geometry

from muninn.enum import Enum
from muninn.exceptions import *
from muninn.function import Prototype, FunctionTable
from muninn.schema import *
from muninn.visitor import Visitor


#
# Table of all supported operators and functions
#
type_map = {
    UUID: Boolean,
}
function_table = FunctionTable(type_map=type_map)
#
# Logical operators
#
function_table.add(Prototype("not", (Boolean,), Boolean))
function_table.add(Prototype("and", (Boolean, Boolean), Boolean))
function_table.add(Prototype("or", (Boolean, Boolean), Boolean))
#
# Membership operators
#
function_table.add(Prototype("in", (Integer, Sequence), Boolean))
function_table.add(Prototype("in", (Long, Sequence), Boolean))
function_table.add(Prototype("in", (Real, Sequence), Boolean))
function_table.add(Prototype("in", (Text, Sequence), Boolean))
function_table.add(Prototype("not in", (Integer, Sequence), Boolean))
function_table.add(Prototype("not in", (Long, Sequence), Boolean))
function_table.add(Prototype("not in", (Real, Sequence), Boolean))
function_table.add(Prototype("not in", (Text, Sequence), Boolean))
#
# Comparison operators
#
function_table.add(Prototype("==", (Long, Long), Boolean))
function_table.add(Prototype("==", (Long, Integer), Boolean))
function_table.add(Prototype("==", (Integer, Long), Boolean))
function_table.add(Prototype("==", (Integer, Integer), Boolean))
function_table.add(Prototype("==", (Real, Real), Boolean))
function_table.add(Prototype("==", (Real, Long), Boolean))
function_table.add(Prototype("==", (Long, Real), Boolean))
function_table.add(Prototype("==", (Real, Integer), Boolean))
function_table.add(Prototype("==", (Integer, Real), Boolean))
function_table.add(Prototype("==", (Boolean, Boolean), Boolean))
function_table.add(Prototype("==", (Text, Text), Boolean))
function_table.add(Prototype("==", (Timestamp, Timestamp), Boolean))
function_table.add(Prototype("==", (UUID, UUID), Boolean))

function_table.add(Prototype("!=", (Long, Long), Boolean))
function_table.add(Prototype("!=", (Long, Integer), Boolean))
function_table.add(Prototype("!=", (Integer, Long), Boolean))
function_table.add(Prototype("!=", (Integer, Integer), Boolean))
function_table.add(Prototype("!=", (Real, Real), Boolean))
function_table.add(Prototype("!=", (Real, Long), Boolean))
function_table.add(Prototype("!=", (Long, Real), Boolean))
function_table.add(Prototype("!=", (Real, Integer), Boolean))
function_table.add(Prototype("!=", (Integer, Real), Boolean))
function_table.add(Prototype("!=", (Boolean, Boolean), Boolean))
function_table.add(Prototype("!=", (Text, Text), Boolean))
function_table.add(Prototype("!=", (Timestamp, Timestamp), Boolean))
function_table.add(Prototype("!=", (UUID, UUID), Boolean))

function_table.add(Prototype("<", (Long, Long), Boolean))
function_table.add(Prototype("<", (Long, Integer), Boolean))
function_table.add(Prototype("<", (Integer, Long), Boolean))
function_table.add(Prototype("<", (Integer, Integer), Boolean))
function_table.add(Prototype("<", (Real, Real), Boolean))
function_table.add(Prototype("<", (Real, Long), Boolean))
function_table.add(Prototype("<", (Long, Real), Boolean))
function_table.add(Prototype("<", (Real, Integer), Boolean))
function_table.add(Prototype("<", (Integer, Real), Boolean))
function_table.add(Prototype("<", (Text, Text), Boolean))
function_table.add(Prototype("<", (Timestamp, Timestamp), Boolean))

function_table.add(Prototype(">", (Long, Long), Boolean))
function_table.add(Prototype(">", (Long, Integer), Boolean))
function_table.add(Prototype(">", (Integer, Long), Boolean))
function_table.add(Prototype(">", (Integer, Integer), Boolean))
function_table.add(Prototype(">", (Real, Real), Boolean))
function_table.add(Prototype(">", (Real, Long), Boolean))
function_table.add(Prototype(">", (Long, Real), Boolean))
function_table.add(Prototype(">", (Real, Integer), Boolean))
function_table.add(Prototype(">", (Integer, Real), Boolean))
function_table.add(Prototype(">", (Text, Text), Boolean))
function_table.add(Prototype(">", (Timestamp, Timestamp), Boolean))

function_table.add(Prototype("<=", (Long, Long), Boolean))
function_table.add(Prototype("<=", (Long, Integer), Boolean))
function_table.add(Prototype("<=", (Integer, Long), Boolean))
function_table.add(Prototype("<=", (Integer, Integer), Boolean))
function_table.add(Prototype("<=", (Real, Real), Boolean))
function_table.add(Prototype("<=", (Real, Long), Boolean))
function_table.add(Prototype("<=", (Long, Real), Boolean))
function_table.add(Prototype("<=", (Real, Integer), Boolean))
function_table.add(Prototype("<=", (Integer, Real), Boolean))
function_table.add(Prototype("<=", (Text, Text), Boolean))
function_table.add(Prototype("<=", (Timestamp, Timestamp), Boolean))

function_table.add(Prototype(">=", (Long, Long), Boolean))
function_table.add(Prototype(">=", (Long, Integer), Boolean))
function_table.add(Prototype(">=", (Integer, Long), Boolean))
function_table.add(Prototype(">=", (Integer, Integer), Boolean))
function_table.add(Prototype(">=", (Real, Real), Boolean))
function_table.add(Prototype(">=", (Real, Long), Boolean))
function_table.add(Prototype(">=", (Long, Real), Boolean))
function_table.add(Prototype(">=", (Real, Integer), Boolean))
function_table.add(Prototype(">=", (Integer, Real), Boolean))
function_table.add(Prototype(">=", (Text, Text), Boolean))
function_table.add(Prototype(">=", (Timestamp, Timestamp), Boolean))

function_table.add(Prototype("~=", (Text, Text), Boolean))

function_table.add(Prototype("+", (Long,), Long))
function_table.add(Prototype("+", (Integer,), Integer))
function_table.add(Prototype("+", (Real,), Real))

function_table.add(Prototype("-", (Long,), Long))
function_table.add(Prototype("-", (Integer,), Integer))
function_table.add(Prototype("-", (Real,), Real))

function_table.add(Prototype("+", (Long, Long), Long))
function_table.add(Prototype("+", (Long, Integer), Long))
function_table.add(Prototype("+", (Integer, Long), Long))
function_table.add(Prototype("+", (Integer, Integer), Integer))
function_table.add(Prototype("+", (Real, Real), Real))
function_table.add(Prototype("+", (Real, Long), Real))
function_table.add(Prototype("+", (Long, Real), Real))
function_table.add(Prototype("+", (Real, Integer), Real))
function_table.add(Prototype("+", (Integer, Real), Real))

function_table.add(Prototype("-", (Long, Long), Long))
function_table.add(Prototype("-", (Long, Integer), Long))
function_table.add(Prototype("-", (Integer, Long), Long))
function_table.add(Prototype("-", (Integer, Integer), Integer))
function_table.add(Prototype("-", (Real, Real), Real))
function_table.add(Prototype("-", (Real, Long), Real))
function_table.add(Prototype("-", (Long, Real), Real))
function_table.add(Prototype("-", (Real, Integer), Real))
function_table.add(Prototype("-", (Integer, Real), Real))

function_table.add(Prototype("*", (Long, Long), Long))
function_table.add(Prototype("*", (Long, Integer), Long))
function_table.add(Prototype("*", (Integer, Long), Long))
function_table.add(Prototype("*", (Integer, Integer), Integer))
function_table.add(Prototype("*", (Real, Real), Real))
function_table.add(Prototype("*", (Real, Long), Real))
function_table.add(Prototype("*", (Long, Real), Real))
function_table.add(Prototype("*", (Real, Integer), Real))
function_table.add(Prototype("*", (Integer, Real), Real))

function_table.add(Prototype("/", (Long, Long), Long))
function_table.add(Prototype("/", (Long, Integer), Long))
function_table.add(Prototype("/", (Integer, Long), Long))
function_table.add(Prototype("/", (Integer, Integer), Integer))
function_table.add(Prototype("/", (Real, Real), Real))
function_table.add(Prototype("/", (Real, Long), Real))
function_table.add(Prototype("/", (Long, Real), Real))
function_table.add(Prototype("/", (Real, Integer), Real))
function_table.add(Prototype("/", (Integer, Real), Real))

function_table.add(Prototype("-", (Timestamp, Timestamp), Real))

#
# Functions.
#
function_table.add(Prototype("covers", (Geometry, Geometry), Boolean))
function_table.add(Prototype("covers", (Timestamp, Timestamp, Timestamp, Timestamp), Boolean))
function_table.add(Prototype("distance", (Geometry, Geometry), Real))
function_table.add(Prototype("intersects", (Geometry, Geometry), Boolean))
function_table.add(Prototype("intersects", (Timestamp, Timestamp, Timestamp, Timestamp), Boolean))
function_table.add(Prototype("is_defined", (Long,), Boolean))
function_table.add(Prototype("is_defined", (Integer,), Boolean))
function_table.add(Prototype("is_defined", (Real,), Boolean))
function_table.add(Prototype("is_defined", (Boolean,), Boolean))
function_table.add(Prototype("is_defined", (Text,), Boolean))
function_table.add(Prototype("is_defined", (Namespace,), Boolean))
function_table.add(Prototype("is_defined", (Timestamp,), Boolean))
function_table.add(Prototype("is_defined", (UUID,), Boolean))
function_table.add(Prototype("is_defined", (Geometry,), Boolean))
function_table.add(Prototype("is_source_of", (UUID,), Boolean))
function_table.add(Prototype("is_source_of", (Boolean,), Boolean))
function_table.add(Prototype("is_derived_from", (UUID,), Boolean))
function_table.add(Prototype("is_derived_from", (Boolean,), Boolean))
function_table.add(Prototype("has_tag", (Text,), Boolean))
function_table.add(Prototype("now", (), Timestamp))


class TokenType(Enum):
    _items = ("TEXT", "UUID", "TIMESTAMP", "REAL", "INTEGER", "BOOLEAN", "NAME", "OPERATOR", "END")


class Token(object):
    def __init__(self, type_, value=None):
        self.type_ = type_
        self.value = value

    def __repr__(self):
        return "Token(type_ = TokenType.%s, value = %r)" % (TokenType.to_string(self.type_), self.value)


class TokenStream(object):
    _sub_patterns = \
        (
            r"""\"(?:[^\\"]|\\.)*\"""",                                      # Text literals
            r"""\d{4}-\d{2}-\d{2}(?:T\d{2}:\d{2}:\d{2}(?:\.\d{0,6})?)?""",   # Timestamp literals
            r"""[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}""",  # UUID literals
            r"""\d+(?:\.\d*(?:[eE][+-]?\d+)?|[eE][+-]?\d+)""",               # Real literals
            r"""0x[0-9a-fA-F]+|0o[0-7]\d+|0b[0-1]+|\d+""",                   # Integer literals
            r"""<=|>=|==|!=|~=|not in|[*<>@()\[\],.+-/]""",                  # Operators and delimiters
            r"""[a-zA-Z]\w*""",                                              # Names (incl. true, false, in)
        )

    _pattern = r"""(?:%s)""" % ("|".join(["(%s)" % sub_pattern for sub_pattern in _sub_patterns]))
    _re_token = re.compile(_pattern)

    _re_datemin = re.compile(r"0000-00-00(?:T00:00:00(?:\.0{0,6})?)?$")
    _re_datemax = re.compile(r"9999-99-99(?:T99:99:99(?:\.9{0,6})?)?$")

    def __init__(self, text):
        self.text = text
        self.at_end = not self.text
        self.token_start_position, self.token_end_position = 0, 0
        self.next()

    def next(self):
        if self.at_end:
            raise Error("char %d: unexpected end of input" % (self.token_start_position + 1))

        self.token = self._next_token()
        return self.token

    def test(self, types, values=None):
        return False if not self._test_token_types(types) else (values is None or self._test_token_values(values))

    def accept(self, types, values=None):
        if not self.test(types, values):
            return False

        self.next()
        return True

    def expect(self, types, values=None):
        if not self.test(types, values):
            if self.token.type_ == TokenType.END:
                raise Error("char %d: unexpected end of input" % (self.token_start_position + 1))
            else:
                if self.token.value is None:
                    token_str = TokenType.to_string(self.token.type_)
                else:
                    token_str = "\"%s\"" % self.token.value

                expected_str = self._types_to_string(types) if values is None else self._values_to_string(values)
                raise Error("char %d: expected %s, got %s" % (self.token_start_position + 1, expected_str, token_str))

        token = self.token
        self.next()
        return token

    def _types_to_string(self, types):
        try:
            strings = [TokenType.to_string(t) for t in types]
        except TypeError:
            return TokenType.to_string(types)

        return "%s%s" % ("" if len(strings) == 1 else "one of: ", ", ".join(strings))

    def _values_to_string(self, values):
        if isinstance(values, basestring):
            return "\"%s\"" % values

        try:
            strings = ["\"%s\"" % value for value in values]
        except TypeError:
            return "\"%s\"" % values

        return "%s%s" % ("" if len(strings) == 1 else "one of: ", ", ".join(strings))

    def _test_token_types(self, types):
        try:
            return self.token.type_ in types
        except TypeError:
            return self.token.type_ == types

    def _test_token_values(self, values):
        if isinstance(values, basestring):
            return self.token.value == values

        try:
            return self.token.value in values
        except TypeError:
            return self.token.value == values

    def _next_token(self):
        self.token_start_position = self._skip_white_space(self.token_end_position)

        if self.token_start_position == len(self.text):
            self.at_end = True
            return Token(TokenType.END)

        match_object = self._re_token.match(self.text, self.token_start_position)
        if match_object is None:
            raise Error("char %d: syntax error: \"%s\"" % (self.token_start_position + 1,
                        self.text[self.token_start_position:]))

        self.token_start_position, self.token_end_position = match_object.span()
        text, timestamp, uuid_, real, integer, operator, name = match_object.groups()

        if text is not None:
            return Token(TokenType.TEXT, string_unescape(text[1:-1]))

        if timestamp is not None:
            return Token(TokenType.TIMESTAMP, self._parse_timestamp(timestamp))

        if uuid_ is not None:
            return Token(TokenType.UUID, uuid.UUID(uuid_))

        if real is not None:
            return Token(TokenType.REAL, float(real))

        if integer is not None:
            if integer.startswith('0x'):
                base = 16
            elif integer.startswith('0o'):
                base = 8
            elif integer.startswith('0b'):
                base = 2
            else:
                base = 10
            return Token(TokenType.INTEGER, int(integer, base))

        if operator is not None:
            return Token(TokenType.OPERATOR, operator)

        if name is not None:
            if name in ["true", "false"]:
                return Token(TokenType.BOOLEAN, name == "true")
            elif name == "in":
                return Token(TokenType.OPERATOR, name)
            else:
                return Token(TokenType.NAME, name)

        raise Error("char %d: syntax error: \"%s\"" % (self.token_start_position + 1, match_object.group()))

    def _skip_white_space(self, start):
        while start < len(self.text) and self.text[start].isspace():
            start += 1
        return start

    def _parse_timestamp(self, timestamp):
        if self._re_datemin.match(timestamp) is not None:
            return datetime.datetime.min

        if self._re_datemax.match(timestamp) is not None:
            return datetime.datetime.max

        for format_string in ("%Y-%m-%d", "%Y-%m-%dT%H:%M:%S", "%Y-%m-%dT%H:%M:%S.%f"):
            try:
                return datetime.datetime.strptime(timestamp, format_string)
            except ValueError:
                pass

        raise Error("char %d: invalid timestamp: \"%s\"" % (self.token_start_position + 1, timestamp))


class AbstractSyntaxTreeNode(object):
    pass


class Literal(AbstractSyntaxTreeNode):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "(%s %s)" % (type(self).__name__, self.value)


class Name(AbstractSyntaxTreeNode):
    def __init__(self, value):
        self.value = value

    def __str__(self):
        return "(%s %s)" % (type(self).__name__, self.value)


class List(AbstractSyntaxTreeNode):
    def __init__(self, values):
        self.values = values

    def __str__(self):
        return "(%s %s)" % (type(self).__name__, self.value)


class ParameterReference(AbstractSyntaxTreeNode):
    def __init__(self, name):
        self.name = name

    def __str__(self):
        return "(%s %s)" % (type(self).__name__, self.name)


class FunctionCall(AbstractSyntaxTreeNode):
    def __init__(self, name, *args):
        self.name = name
        self.arguments = list(args)

    def __str__(self):
        if not self.arguments:
            return "(%s %s)" % (type(self).__name__, self.name)
        return "(%s %s %s)" % (type(self).__name__, self.name, " ".join(map(str, self.arguments)))


def parse_sequence(stream, parse_item_function, start='(', end=')'):
    stream.expect(TokenType.OPERATOR, start)
    if stream.accept(TokenType.OPERATOR, end):
        return []

    sequence = [parse_item_function(stream)]
    while stream.accept(TokenType.OPERATOR, ","):
        sequence.append(parse_item_function(stream))
    stream.expect(TokenType.OPERATOR, end)
    return sequence


def parse_geometry_sequence(stream, parse_item_function):
    if stream.accept(TokenType.NAME, "EMPTY"):
        return []

    stream.expect(TokenType.OPERATOR, "(")
    sequence = [parse_item_function(stream)]
    while stream.accept(TokenType.OPERATOR, ","):
        sequence.append(parse_item_function(stream))
    stream.expect(TokenType.OPERATOR, ")")
    return sequence


def parse_signed_coordinate(stream):
    if stream.accept(TokenType.OPERATOR, "-"):
        token = stream.expect((TokenType.INTEGER, TokenType.REAL))
        return -float(token.value)

    stream.accept(TokenType.OPERATOR, "+")
    token = stream.expect((TokenType.INTEGER, TokenType.REAL))
    return float(token.value)


def parse_point_raw(stream):
    return geometry.Point(parse_signed_coordinate(stream),
                          parse_signed_coordinate(stream))


def parse_point(stream):
    stream.expect(TokenType.OPERATOR, "(")
    point = parse_point_raw(stream)
    stream.expect(TokenType.OPERATOR, ")")
    return point


def parse_line_string(stream):
    return geometry.LineString(parse_geometry_sequence(stream, parse_point_raw))


def parse_linear_ring(stream):
    points = parse_geometry_sequence(stream, parse_point_raw)
    if len(points) == 0:
        return geometry.LinearRing()

    if len(points) < 4:
        raise Error("char %d: linear ring should be empty or should contain >= 4 points" % stream.token_start_position)

    if points[-1] != points[0]:
        raise Error("char %d: linear ring should be closed" % stream.token_start_position)

    return geometry.LinearRing(points[:-1])


def parse_polygon(stream):
    return geometry.Polygon(parse_geometry_sequence(stream, parse_linear_ring))


def parse_multi_point(stream):
    return geometry.MultiPoint(parse_geometry_sequence(stream, parse_point))


def parse_multi_line_string(stream):
    return geometry.MultiLineString(parse_geometry_sequence(stream, parse_line_string))


def parse_multi_polygon(stream):
    return geometry.MultiPolygon(parse_geometry_sequence(stream, parse_polygon))


def parse_atom(stream):
    # Sub-expression.
    if stream.accept(TokenType.OPERATOR, "("):
        sub_expression = parse_expression(stream)
        stream.expect(TokenType.OPERATOR, ")")
        return sub_expression

    # Parameter reference.
    if stream.accept(TokenType.OPERATOR, "@"):
        name_token = stream.expect(TokenType.NAME)
        return ParameterReference(name_token.value)

    # Geometry literal, function call, or name.
    if stream.test(TokenType.NAME):
        name_token = stream.expect(TokenType.NAME)

        # Geometry literals.
        if name_token.value == "POINT":
            return Literal(parse_point(stream))
        elif name_token.value == "LINESTRING":
            return Literal(parse_line_string(stream))
        elif name_token.value == "POLYGON":
            return Literal(parse_polygon(stream))
        elif name_token.value == "MULTIPOINT":
            return Literal(parse_multi_point(stream))
        elif name_token.value == "MULTILINESTRING":
            return Literal(parse_multi_line_string(stream))
        elif name_token.value == "MULTIPOLYGON":
            return Literal(parse_multi_polygon(stream))

        # Function call.
        if stream.test(TokenType.OPERATOR, "("):
            return FunctionCall(name_token.value, *parse_sequence(stream, parse_expression))

        # Name (possibly qualified).
        parts = [name_token.value]
        while stream.accept(TokenType.OPERATOR, "."):
            name_token = stream.expect(TokenType.NAME)
            parts.append(name_token.value)
        return Name(".".join(parts))

    if stream.test(TokenType.OPERATOR, "["):
        return List(parse_sequence(stream, parse_expression, "[", "]"))

    # Literal.
    token = stream.expect((TokenType.TEXT, TokenType.TIMESTAMP, TokenType.UUID, TokenType.REAL, TokenType.INTEGER,
                           TokenType.BOOLEAN))
    return Literal(token.value)


def parse_term(stream):
    if stream.test(TokenType.OPERATOR, ("+", "-")):
        operator_token = stream.expect(TokenType.OPERATOR, ("+", "-"))
        return FunctionCall(operator_token.value, parse_term(stream))
    return parse_atom(stream)


def parse_arithmetic_expression(stream):
    lhs = parse_term(stream)
    if stream.test(TokenType.OPERATOR, ("+", "-", "*", "/")):
        operator_token = stream.expect(TokenType.OPERATOR, ("+", "-", "*", "/"))
        return FunctionCall(operator_token.value, lhs, parse_arithmetic_expression(stream))
    return lhs


def parse_comparison(stream):
    lhs = parse_arithmetic_expression(stream)
    if stream.test(TokenType.OPERATOR, ("<", ">", "==", ">=", "<=", "!=", "~=", "in", "not in")):
        operator_token = stream.expect(TokenType.OPERATOR, ("<", ">", "==", ">=", "<=", "!=", "~=", "in", "not in"))
        return FunctionCall(operator_token.value, lhs, parse_comparison(stream))
    return lhs


def parse_not_expression(stream):
    if stream.accept(TokenType.NAME, "not"):
        return FunctionCall("not", parse_not_expression(stream))
    return parse_comparison(stream)


def parse_and_expression(stream):
    lhs = parse_not_expression(stream)
    if stream.accept(TokenType.NAME, "and"):
        return FunctionCall("and", lhs, parse_and_expression(stream))
    return lhs


def parse_or_expression(stream):
    lhs = parse_and_expression(stream)
    if stream.accept(TokenType.NAME, "or"):
        return FunctionCall("or", lhs, parse_or_expression(stream))
    return lhs


def parse_expression(stream):
    return parse_or_expression(stream)


def _literal_type(literal):
    if isinstance(literal, list):  # TODO use Sequence.validate..?
        return Sequence

    for type in (Text, Timestamp, UUID, Boolean, Integer, Long, Real, Geometry):
        try:
            type.validate(literal)
        except ValueError:
            pass
        else:
            return type

    raise Error("unable to determine type of literal value: %r" % literal)


class SemanticAnalysis(Visitor):
    def __init__(self, namespace_schemas, parameters, having=False):
        super(SemanticAnalysis, self).__init__()
        self._namespace_schemas = namespace_schemas
        self._parameters = parameters
        self._having = having

    def visit_Literal(self, visitable):
        visitable.type = _literal_type(visitable.value)

    def visit_Name(self, visitable):
        if self._having:
            item = Identifier(visitable.value, self._namespace_schemas)
            visitable.type = item.muninn_type
            visitable.value = item
            return

        split_name = visitable.value.split(".")

        # namespace/implicit core property
        if len(split_name) == 1:
            if split_name[0] in self._namespace_schemas:
                namespace = split_name[0]
                name = None
            else:
                namespace, name = "core", split_name[0]

        # namespace.property
        elif len(split_name) == 2:
            namespace, name = split_name

        else:
            raise Error("invalid property name: \"%s\"" % visitable.value)

        # check that namespace exists
        try:
            schema = self._namespace_schemas[namespace]
        except KeyError:
            raise Error("undefined namespace: \"%s\"" % namespace)

        # namespace
        if name is None:
            visitable.value = split_name[0]
            visitable.type = Namespace

        # namespace.property
        else:
            try:
                type_ = schema[name]
            except KeyError:
                if len(split_name) == 2:
                    raise Error("undefined property: \"%s\"" % visitable.value)
                else:
                    raise Error("undefined name: \"%s\"" % name)

            visitable.value = "%s.%s" % (namespace, name)
            visitable.type = type_

    def visit_List(self, visitable):  # TODO check same literal type
        values = []
        for value in visitable.values:
            if not isinstance(value, Literal):
                raise Error("list contains non-literal")
            values.append(value.value)
        visitable.value = values
        visitable.type = Sequence

    def visit_ParameterReference(self, visitable):
        try:
            value = self._parameters[visitable.name]
        except KeyError:
            raise Error("no value for parameter: \"%s\"" % visitable.name)

        visitable.value = value
        visitable.type = _literal_type(value)

    def visit_FunctionCall(self, visitable):
        # Resolve the type of the function arguments.
        for argument in visitable.arguments:
            self.visit(argument)

        prototype = Prototype(visitable.name, [argument.type for argument in visitable.arguments])

        try:
            prototypes = function_table.resolve(prototype)
        except KeyError:
            prototypes = []

        if not prototypes:
            raise Error("undefined function: \"%s\"" % prototype)

        if len(prototypes) > 1:
            raise InternalError("cannot uniquely resolve function: \"%s\"" % prototype)

        prototype = prototypes[0]
        visitable.prototype = prototype
        visitable.type = prototype.return_type

    def visit_AbstractSyntaxTreeNode(self, visitable):
        if not hasattr(visitable, "type"):
            raise InternalError("encountered abstract syntax tree node without type attribute: %s" %
                                type(visitable).__name__)

    def default(self, visitable):
        raise InternalError("unsupported abstract syntax tree node type: %s" % type(visitable).__name__)


def parse(text):
    stream = TokenStream(text)
    abstract_syntax_tree = parse_expression(stream)
    if not stream.test(TokenType.END):
        raise Error("char %d: extra characters after expression: \"%s\"" % (stream.token_start_position + 1,
                                                                            text[stream.token_start_position:]))
    return abstract_syntax_tree


def analyze(abstract_syntax_tree, namespace_schemas={}, parameters={}, having=False):
    annotated_syntax_tree = copy.deepcopy(abstract_syntax_tree)
    SemanticAnalysis(namespace_schemas, parameters, having=having).visit(annotated_syntax_tree)
    return annotated_syntax_tree


def parse_and_analyze(text, namespace_schemas={}, parameters={}, having=False):
    return analyze(parse(text), namespace_schemas, parameters, having)


def string_unescape(text):
    '''
    Unescape special characters in a string.
    Python2 and 3 compatible, uses the native string type.
    In python2, the same effect can also be achieved with `string.decode("string-escape")`
    '''
    text = str(text)  # make sure we are using the native string type
    regex = re.compile('\\\\(\\\\|[\'"abfnrtv])')
    translator = {
        '\\': '\\',
        "'": "'",
        '"': '"',
        'a': '\a',
        'b': '\b',
        'f': '\f',
        'n': '\n',
        'r': '\r',
        't': '\t',
        'v': '\v',
    }

    def _replace(m):
        c = m.group(1)
        return translator[c]

    result = regex.sub(_replace, text)
    return result


class Identifier(object):
    def __init__(self, canonical_identifier, namespace_schemas):
        self.canonical = canonical_identifier
        self.subscript = None

        if canonical_identifier == 'tag':
            # the rules to get the namespace database table name also apply to 'tag'
            self.namespace = canonical_identifier
            self.identifier = canonical_identifier
            self.muninn_type = Text

        elif canonical_identifier == 'count':
            self.namespace = None
            self.identifier = canonical_identifier
            self.muninn_type = Long

        elif not re.match(r'[\w]+\.[\w.]+', canonical_identifier):
            raise Error("cannot resolve identifier: %r" % canonical_identifier)

        else:
            segments = canonical_identifier.split('.')

            if len(segments) == 1:
                self.namespace = 'core'
                self.identifier = segments[0]

            elif len(segments) == 2:
                if segments[0] in namespace_schemas:
                    self.namespace, self.identifier = segments
                else:
                    self.namespace = 'core'
                    self.identifier, self.subscript = segments

            elif len(segments) == 3:
                self.namespace, self.identifier, self.subscript = segments

            else:
                raise Error("cannot resolve identifier: %r" % canonical_identifier)

            # check if namespace is valid
            if self.namespace not in namespace_schemas:
                raise Error("undefined namespace: \"%s\"" % self.namespace)

            # check if property name is valid
            if self.identifier not in namespace_schemas[self.namespace]:
                if self.property_name != 'core.validity_duration':
                    raise Error("no property: %r defined within namespace: %r" % (self.identifier, self.namespace))

            # note: not checking if subscript is valid; the list of possible subscripts varies depending on context
            if self.property_name == 'core.validity_duration':
                self.muninn_type = None
            else:
                self.muninn_type = namespace_schemas[self.namespace][self.identifier]

    @property
    def property_name(self):
        return '%s.%s' % (self.namespace, self.identifier)

    @property
    def resolve(self):
        if self.canonical in ('count', 'tag'):
            return self.canonical
        elif self.subscript is None:
            return '%s.%s' % (self.namespace, self.identifier)
        else:
            return '%s.%s.%s' % (self.namespace, self.identifier, self.subscript)



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

from muninn.enum import Enum


class CascadeRule(Enum):
    _items = ("IGNORE", "CASCADE_PURGE_AS_STRIP", "CASCADE_PURGE", "STRIP", "CASCADE", "PURGE")



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function

import muninn

from muninn.tools.utils import create_parser, parse_args_and_run


def list_tags(args):
    with muninn.open(args.archive) as archive:
        for product in archive.search(where=args.expression, property_names=['uuid', 'product_name']):
            tags = archive.tags(product.core.uuid)
            print("%s (%s): %s" % (product.core.product_name, product.core.uuid, ", ".join(tags)))

    return 0


def main():
    parser = create_parser(description="List tags of products contained in a muninn archive")
    parser.add_argument("archive", metavar="ARCHIVE", help="identifier of the archive to use")
    parser.add_argument("expression", metavar="EXPRESSION",
                        help="expression used to search for products to list tags of")
    return parse_args_and_run(parser, list_tags)


if __name__ == '__main__':
    main()



#
# Copyright (C) 2014-2024 S[&]T, The Netherlands.
#

from __future__ import absolute_import, division, print_function
from muninn._compat import itervalues, imap, izip

try:
    from collections.abc import MutableSet
except ImportError:
    from collections import MutableSet


class Prototype(object):
    def __init__(self, name, argument_types=(), return_type=None):
        self.name = name
        self.argument_types = argument_types
        self.return_type = return_type

        self._id = "%s(%s)" % (self.name, ",".join(imap(lambda type: type.name(), self.argument_types)))
        if self.return_type is not None:
            self._id += " " + self.return_type.name()

    @property
    def id(self):
        return self._id

    @property
    def arity(self):
        return len(self.argument_types)

    def _argument_types_equal(self, other):
        if self.arity != other.arity:
            return False

        for type_self, type_other in izip(self.argument_types, other.argument_types):
            if type_self is not type_other:
                return False
        return True

    def __eq__(self, other):
        return self.name == other.name and self._argument_types_equal(other) and self.return_type is other.return_type

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return self.id

    def __repr__(self):
        return "Prototype(name=%r, argument_types=%r, return_type=%r)" % (self.name, self.argument_types,
                                                                          self.return_type)

    def __hash__(self):
        return hash(self.id)


class FunctionTable(MutableSet):
    def __init__(self, prototypes=[], type_map=None):
        self._prototypes = {}
        self._type_map = type_map
        for prototype in prototypes:
            self.add(prototype)

    def __contains__(self, prototype):
        try:
            prototypes = self._prototypes[prototype.name]
        except KeyError:
            return False
        else:
            return prototype in prototypes

    def __iter__(self):
        for prototypes in itervalues(self._prototypes):
            for prototype in prototypes:
                yield prototype

    def __len__(self):
        return sum(imap(len, itervalues(self._prototypes)))

    def add(self, prototype):
        try:
            self._prototypes[prototype.name].add(prototype)
        except KeyError:
            self._prototypes[prototype.name] = set((prototype,))

    def discard(self, prototype):
        try:
            self._prototypes[prototype.name].discard(prototype)
        except KeyError:
            pass

    def resolve(self, prototype):
        top, top_equal = [], 0
        for candidate in self._prototypes[prototype.name]:
            if candidate.arity != prototype.arity:
                continue

            equal, compatible = 0, 0
            for candidate_type, type_ in izip(candidate.argument_types, prototype.argument_types):
                if type_ is candidate_type:
                    equal += 1
                elif type_ in self._type_map and self._type_map[type_] is candidate_type:
                    compatible += 1
                elif type_ in self._type_map and issubclass(type_, self._type_map[type_]):
                    compatible += 1
                elif issubclass(type_, candidate_type):
                    compatible += 1
                else:
                    break

            if equal + compatible != prototype.arity:
                continue

            if equal > top_equal:
                top = [candidate]
                top_equal = equal
            elif equal == top_equal:
                top.append(candidate)

        return top
